import sys
import ast
import os
import numpy as np
import scipy
import itertools
import random
from sklearn.metrics.pairwise import cosine_similarity

from graphmb.visualize import run_tsne, plot_embs
from graphmb.utils import run_clustering
# code to run evaluation based on lineage.ms file (Bacteria) and marker_gene_stats.txt file

# Get precicion
def getPrecision(mat, k, s, total):
    sum_k = 0
    for i in range(k):
        max_s = 0
        for j in range(s):
            if mat[i][j] > max_s:
                max_s = mat[i][j]
        sum_k += max_s  
    return sum_k / total


# Get recall
def getRecall(mat, k, s, total, unclassified):
    sum_s = 0
    for i in range(s):
        max_k = 0
        for j in range(k):
            if mat[j][i] > max_k:
                max_k = mat[j][i]
        sum_s += max_k
    return sum_s / (total + unclassified)


def getAveragePrecision(cluster_to_labels, cluster_sizes):
    avg_precision = []
    for c in cluster_to_labels:
        #TP = cluster_to_labels[c][1]; TP+FP + cluster_sizes[c]
        cluster_p = cluster_to_labels[c][1] / cluster_sizes[c]
        avg_precision.append(cluster_p)
    return round(sum(avg_precision) / len(avg_precision), 4)

def getAverageRecall(label_to_cluster, cluster_to_contig, contig_sizes, label_to_node):
    avg_recall = []
    for label in label_to_cluster:
        if label_to_cluster[label][0] == "NA":
            avg_recall.append(0)
        else:
            cluster_r = label_to_cluster[label][1] / sum([contig_sizes.get(n, 1) for n in cluster_to_contig[label_to_cluster[label][0]]])
            avg_recall.append(cluster_r)
    return round(sum(avg_recall) / len(avg_recall), 4)


# Get ARI
def getARI(mat, k, s, N):
    t1 = 0
    for i in range(k):
        sum_k = 0
        for j in range(s):
            sum_k += mat[i][j]
        t1 += scipy.special.binom(sum_k, 2)
    t2 = 0
    for i in range(s):
        sum_s = 0
        for j in range(k):
            sum_s += mat[j][i]
        t2 += scipy.special.binom(sum_s, 2)
    t3 = t1 * t2 / scipy.special.binom(N, 2)
    t = 0
    for i in range(k):
        for j in range(s):
            t += scipy.special.binom(mat[i][j], 2)
    ari = (t - t3) / ((t1 + t2) / 2 - t3)
    return ari


# Get F1-score
def getF1(prec, recall):
    if prec == 0.0 or recall == 0.0:
        return 0.0
    else:
        return 2 * prec * recall / (prec + recall)


def calculate_overall_prf(cluster_to_contig, contig_to_cluster, node_to_label, label_to_node):
    # calculate how many contigs are in the majority class of each cluster
    total_binned = 0
    # convert everything to ids
    labels = list(label_to_node.keys())
    clusters = list(cluster_to_contig.keys())
    n_pred_labels = len(clusters)
    n_true_labels = len(labels)
    ground_truth_count = len(node_to_label)
    # empty matrix
    bins_species = [[0 for x in range(n_true_labels)] for y in range(n_pred_labels)]
    # update with cluster memberships
    for i in contig_to_cluster:
        if i in node_to_label:
            # breakpoint()
            total_binned += 1
            bins_species[clusters.index(contig_to_cluster[i])][labels.index(node_to_label[i])] += 1
    my_precision = getPrecision(bins_species, n_pred_labels, n_true_labels, total_binned)
    my_recall = getRecall(
        bins_species, n_pred_labels, n_true_labels, total_binned, (ground_truth_count - total_binned)
    )
    #my_ari = getARI(bins_species, n_pred_labels, n_true_labels, total_binned)
    my_f1 = getF1(my_precision, my_recall)
    return my_precision, my_recall, my_f1, 0

def calculate_sim_between_same_labels_small(node_names, embeddings, edges, label_to_node, node_to_label):
    # divide sim between node of same label by avg sim between all nodes
    avg_label_sims = {}
    # use only embeddings of nodes with labels
    edges = np.array(edges)
    # calculate sim between all embs
    all_cosine_sim = cosine_similarity(embeddings, embeddings)
    #https://stackoverflow.com/a/69865919
    i = np.ravel_multi_index(np.array(edges).T, all_cosine_sim.shape)
    edge_sim = all_cosine_sim.take(i) # get sim between embs of edges
    edge_sim = edge_sim.sum()
    #all_cosine_sim = np.triu(all_cosine_sim)
    # all sim should not include sims of edges
    all_cosine_sim = all_cosine_sim.sum() - edge_sim
    all_cosine_sim = all_cosine_sim / ((embeddings.shape[0]**2)-len(edges))
    edge_sim /= len(edges)
    for l in label_to_node:
        label_node_idxs = [node_names.index(n) for n in label_to_node[l]]
        label_embs = np.array(embeddings)[label_node_idxs]
        if label_embs.shape[0] > 1: # at least two nodes
            avg_label_sims[l] = ((cosine_similarity(label_embs, label_embs))).mean()
        # pick a random 
    avg = sum(avg_label_sims.values())/len(avg_label_sims.values())      
    #print(round(avg, 4), round(all_cosine_sim, 4))
          #[(x, round(avg_label_sims[x], 4), len(label_to_node[x])) for x in avg_label_sims][:10]])
    return avg, edge_sim, all_cosine_sim


def calculate_sim_between_same_labels_big(node_names, embeddings, edges, label_to_node, node_to_label):
    # divide sim between node of same label by avg sim between all nodes
    avg_label_sims = {}
    # use only embeddings of nodes in edges to make it more efficient
    edges = np.array(edges)
    node_idx_with_edges = np.unique(edges)
    #label_mask = np.array([node_to_label.get(n, "NA") != "NA" for n in node_names])
    all_cosine_sim = cosine_similarity(embeddings[node_idx_with_edges], embeddings[node_idx_with_edges])
    edges_new = np.searchsorted(node_idx_with_edges, edges)
    #https://stackoverflow.com/a/69865919
    i = np.ravel_multi_index(np.array(edges_new).T, all_cosine_sim.shape)
    edge_sim = all_cosine_sim.take(i)
    edge_sim = edge_sim.mean()
    all_cosine_sim = np.triu(all_cosine_sim)
    all_cosine_sim = all_cosine_sim.mean()
    for l in label_to_node:
        label_node_idxs = [node_names.index(n) for n in label_to_node[l] if node_names.index(n) in node_idx_with_edges]
        label_node_idxs = np.searchsorted(node_idx_with_edges, label_node_idxs)
        label_embs = np.array(embeddings)[label_node_idxs]
        if label_embs.shape[0] > 1: # at least two nodes
            avg_label_sims[l] = (np.triu(cosine_similarity(label_embs, label_embs))).mean()
        # pick a random 
    avg = sum(avg_label_sims.values())/len(avg_label_sims.values())      
    #print(round(avg, 4), round(all_cosine_sim, 4))
          #[(x, round(avg_label_sims[x], 4), len(label_to_node[x])) for x in avg_label_sims][:10]])
    return avg, edge_sim, all_cosine_sim


def read_marker_gene_sets(lineage_file):
    """Open file with gene sets for a taxon

    :param lineage_file: path to marker set file from CheckM (Bacteria.ms)
    :type lineage_file: str
    :return: Marker sets
    :rtype: set
    """
    with open(lineage_file, "r") as f:
        lines = f.readlines()
    # consider only single taxon gene set
    sets = lines[1].strip().split("\t")[-1]
    sets = ast.literal_eval(sets)
    return sets


def read_contig_genes(contig_markers):
    """Open file mapping contigs to genes

    :param contig_markers: path to contig markers (marker stats)
    :type contig_markers: str
    :return: Mapping contig names to markers
    :rtype: dict
    """
    contigs = {}
    with open(contig_markers, "r") as f:
        for line in f:
            values = line.strip().split("\t")
            contig_name = values[0]
            # keep only first two elements
            contig_name = "_".join(contig_name.split("_")[:2])
            contigs[contig_name] = {}
            mappings = ast.literal_eval(values[1])
            for contig in mappings:
                for gene in mappings[contig]:
                    if gene not in contigs[contig_name]:
                        contigs[contig_name][gene] = 0
                    # else:
                    #    breakpoint()
                    contigs[contig_name][gene] += 1
                    if len(mappings[contig][gene]) > 1:
                        breakpoint()
    return contigs


def get_markers_to_contigs(marker_sets, contigs):
    """Get marker to contig mapping

    :param marker_sets: Marker sets from CheckM
    :type marker_sets: set
    :param contigs: Contig to marker mapping
    :type contigs: dict
    :return: Marker to contigs list mapping
    :rtype: dict
    """
    marker2contigs = {}
    for marker_set in marker_sets:
        for gene in marker_set:
            marker2contigs[gene] = []
            for contig in contigs:
                if gene in contigs[contig]:
                    marker2contigs[gene].append(contig)
    return marker2contigs


def evaluate_contig_sets(marker_sets, contig_marker_counts, bin_to_contigs):
    """Calculate completeness and contamination for each bin given a set of
       marker gene sets and contig marker counts

    :param marker_sets: reference marker sets, from Bacteria
    :type marker_sets: list
    :param contig_marker_counts: Counts of each gene on each contig
    :type contig_marker_counts: dict
    :param bin_to_contigs: Mapping bins to contigs
    :type bin_to_contigs: dict
    """
    results = {}
    for bin in bin_to_contigs:
        bin_genes = {}
        for contig in bin_to_contigs[bin]:
            if contig not in contig_marker_counts:
                print("missing", contig)
                continue
            for gene in contig_marker_counts[contig]:
                if gene not in bin_genes:
                    bin_genes[gene] = 0
                bin_genes[gene] += contig_marker_counts[contig][gene]
        comp = completeness(marker_sets, set(bin_genes.keys()))
        cont = contamination(marker_sets, bin_genes)
        results[bin] = {"comp": comp, "cont": cont, "genes": bin_genes}
    return results

def completeness(reference_markers, genes):
    numerator = 0.0
    for marker_set in reference_markers:
        common = marker_set & genes
        if len(marker_set) > 0:
            numerator += len(common) / len(marker_set)
    return 100 * (numerator / len(reference_markers))


def contamination(reference_markers, genes):
    numerator = 0.0
    for i, marker_set in enumerate(reference_markers):
        inner_total = 0.0
        for gene in marker_set:
            if gene in genes and genes[gene] > 0:
                inner_total += genes[gene] - 1.0
        if len(marker_set) > 0:
            numerator += inner_total / len(marker_set)
    return 100.0 * (numerator / len(reference_markers))


def compute_cluster_score(reference_markers, contig_genes, node_names, node_labels):
    labels_to_nodes = {i: node_names[node_labels == i].tolist() for i in np.unique(node_labels)}
    results = {}
    for label in labels_to_nodes:
        genes = {}
        for node_name in labels_to_nodes[label]:
            if node_name not in contig_genes:
                # print("missing", node_name)
                continue
            for gene in contig_genes[node_name]:
                if gene not in genes:
                    genes[gene] = 0
                genes[gene] += contig_genes[node_name][gene]

        comp = completeness(reference_markers, set(genes.keys()))
        cont = contamination(reference_markers, genes)
        results[label] = {"comp": comp, "cont": cont, "genes": genes}
    return results


def compute_hq(cluster_stats, comp_th=90, cont_th=5):
    hq = 0
    positive_clusters = []
    for label in cluster_stats:
        if cluster_stats[label]["comp"] >= comp_th and cluster_stats[label]["cont"] < cont_th:
            hq += 1
            positive_clusters.append(label)
    return hq, positive_clusters

def compute_unresolved(reference_markers, contig_genes, node_names, node_labels, resolved_clusters):
    # completeness of a bin with all not HQ -> how many more bins we could get
    unresolved_contigs = np.array([n for i,n in enumerate(node_names) if node_labels[i] not in resolved_clusters])
    cluster_stats = compute_cluster_score(reference_markers, contig_genes,
                                          unresolved_contigs, np.ones(len(unresolved_contigs)))
    contamination = cluster_stats[1]["cont"]
    potential_mags = int(contamination / 100)
    return potential_mags, unresolved_contigs

def compute_clusters_and_stats(
    X,
    node_names,
    dataset,
    k=0,
    clustering="vamb",
    cuda=False,
    tsne=False,
    tsne_path=None,
    max_pos_pairs=None,
    use_labels=False,
    amber=False,
    compute_pospairs=False,
    unresolved=True
):
    reference_markers = dataset.ref_marker_sets
    contig_genes = dataset.contig_markers
    node_to_gt_idx_label = dataset.node_to_label
    gt_idx_label_to_node = dataset.label_to_node
    cluster_to_contig, contig_to_bin, labels, cluster_centroids = run_clustering(X, node_names,
                                                                                 clustering, cuda, k=k)
   
    scores = {"precision": 0, "recall": 0, "f1": 0, "ari": 0, "hq": 0, "mq": 0,
        "n_clusters": len(np.unique(labels)), "unresolved_mags": 0}
    positive_pairs = []
    positive_clusters = []
    if contig_genes is not None and len(contig_genes) > 0:
        cluster_stats = compute_cluster_score(reference_markers, contig_genes, node_names, labels)
        hq, positive_clusters = compute_hq(cluster_stats)
        mq, _ = compute_hq(cluster_stats, comp_th=50,cont_th=10)
        """non_comp, _ = compute_hq(cluster_stats, comp_th=0, cont_th=10)
        all_cont, _ = compute_hq(cluster_stats, comp_th=90, cont_th=1000)
        """
        scores["hq"] = hq
        scores["mq"] = mq
        #TODO avg comp and cont of hq bins
        scores["hq_comp"] = np.mean([cluster_stats[c]["comp"] for c in positive_clusters])
        scores["hq_cont"] = np.mean([cluster_stats[c]["cont"] for c in positive_clusters])
        
        if unresolved:
            unresolved_mags, unresolved_contigs = compute_unresolved(reference_markers=reference_markers,
                contig_genes=contig_genes,
                node_names=node_names,
                node_labels=labels,
                resolved_clusters=positive_clusters)
            scores["unresolved_mags"] = unresolved_mags
            scores["unresolved_contigs"] = len(unresolved_contigs)
            unresolved_contigs_with_scgs = np.array([n for i,n in enumerate(node_names) \
                if labels[i] not in positive_clusters and len(dataset.contig_markers[n]) > 0])
            scores["unresolved_contigs_with_scgs"] = len(unresolved_contigs_with_scgs)
        # print(hq, mq, "incompete but non cont:", non_comp, "cont but complete:", all_cont)
        positive_pairs = None
        if compute_pospairs:
            positive_pairs = get_positive_pairs(node_names, positive_clusters, cluster_to_contig, max_pos_pairs)
    
    # calculate edge similarity metrics
    if node_to_gt_idx_label is not None and len(dataset.labels) > 1:
        #if len(dataset.node_names) < 10_000:
        #    sims = calculate_sim_between_same_labels_small(dataset.node_names, X,
        #                                            list(zip(dataset.edges_src, dataset.edges_dst)),
        #                                            dataset.label_to_node, dataset.node_to_label)
        #elif len([n for n in dataset.node_names if dataset.node_to_label.get(n, "NA") != "NA"]) < 10_000:
        #    sims = calculate_sim_between_same_labels_big(dataset.node_names, X,
        #                                        list(zip(dataset.edges_src, dataset.edges_dst)),
        #                                        dataset.label_to_node, dataset.node_to_label)
        #else:
        sims = 1,1,1
        scores["avg_label_sim"], scores["avg_edge_sim"], scores["avg_total_sim"] = sims
        scores["ratio_labelsim"] = scores["avg_label_sim"]/scores["avg_total_sim"]
        scores["ratio_edgesim"] = scores["avg_edge_sim"]/scores["avg_total_sim"]
        scores["ratio_labeltoedgesim"] = scores["avg_label_sim"]/scores["avg_edge_sim"]

    if amber == True:
        # TODO use p/r/ to get positive_clusters
        import graphmb.amber_eval as amber
        # save to file
        output_bins_filename = dataset.cache_dir + f"/{dataset.name}_temp_contig2bin.tsv"
        amber.write_amber_bins(contig_to_bin, output_bins_filename)
        amber_metrics, bin_counts = amber.amber_eval(os.path.join(dataset.data_dir, dataset.labelsfile), output_bins_filename, ["graphmb"])
        scores["precision_avg_bp"] = amber_metrics["precision_avg_bp"]
        scores["recall_avg_bp"] = amber_metrics["recall_avg_bp"]
        scores["f1_avg_bp"] = amber_metrics["f1_avg_bp"]
        scores["amber_hq"] = bin_counts["> 90% completeness"][1]
        scores["amber_mq"] = bin_counts["> 50% completeness"][0]
    else:
        # calculate edge metrics
        p, r, f1, ari = calculate_overall_prf(
            cluster_to_contig, contig_to_bin, node_to_gt_idx_label, gt_idx_label_to_node
        )
        scores["precision"] = p
        scores["recall"] = r
        scores["f1"] = f1
        scores["ari"] = ari
    #plot tSNE
    if tsne:
        cluster_to_contig = {cluster: [dataset.node_names[i] for i,x in enumerate(labels) if x == cluster] for cluster in set(labels)}
        node_embeddings_2dim, centroids_2dim = run_tsne(X, dataset, cluster_to_contig, positive_clusters, centroids=cluster_centroids)
        plot_embs(
            dataset.node_names,
            node_embeddings_2dim,
            #dataset.label_to_node.copy(),
            cluster_to_contig,
            centroids=centroids_2dim,
            hq_centroids=positive_clusters,
            node_sizes=None,
            outputname=tsne_path,
        )
    #print("calc metrics time", datetime.datetime.now() - processing_time)
    return (
        labels,
        scores,
        positive_pairs,
        positive_clusters,
    )

def get_positive_pairs(node_names, positive_clusters, cluster_to_contig, max_pos_pairs=None):
    positive_pairs = []
    node_names_to_idx = {node_name: i for i, node_name in enumerate(node_names)}
    for label in positive_clusters:
        added_pairs = []
        for (p1, p2) in itertools.combinations(cluster_to_contig[label], 2):
            added_pairs.append((node_names_to_idx[p1], node_names_to_idx[p2]))    
        if max_pos_pairs is not None and len(added_pairs) > max_pos_pairs:
            added_pairs = random.sample(added_pairs, max_pos_pairs)
        positive_pairs.extend(added_pairs)
    # print("found {} positive pairs".format(len(positive_pairs)))
    positive_pairs = np.unique(np.array(list(positive_pairs)), axis=0)
    return positive_pairs

def eval_epoch(logger, summary_writer, node_new_features, cluster_mask, weights,
               step, args, dataset, epoch, scores, best_metric, best_embs, best_epoch, best_model, target_metric="hq"):
    # used only by vgae model
    #log_to_tensorboard(summary_writer, {"Embs average": np.mean(node_new_features), 'Embs std': np.std(node_new_features) }, step)
    cluster_labels, stats, _, hq_bins = compute_clusters_and_stats(
        node_new_features[cluster_mask], np.array(dataset.node_names)[cluster_mask],
        dataset, clustering=args.clusteringalgo, k=args.kclusters, tsne=args.tsne,
        amber=(args.labels is not None and "amber" in args.labels), cuda=args.cuda,
        compute_pospairs=False, unresolved=True
    )
    
    stats["epoch"] = epoch
    scores.append(stats)
    #logger.info(str(stats))

    #log_to_tensorboard(summary_writer, {"hq": stats["hq"], "mq": stats["mq"]}, step)
    #all_cluster_labels.append(cluster_labels)
    #if dataset.contig_markers is not None and len(dataset.contig_markers) > 0 and stats["hq"] > best_metric:
    #    best_metric, best_embs, best_epoch, best_model = stats["hq"], node_new_features, epoch, weights
        #best_model = th.gnn_model
        #save_model(args, e, th, th_vae)
    if stats[target_metric] > best_metric:
        best_metric, best_embs, best_epoch, best_model = stats[target_metric], node_new_features, epoch, weights

    return best_metric, best_embs, best_epoch, scores, best_model, cluster_labels


def eval_epoch_cluster(logger, summary_writer, node_new_features, cluster_mask, best_hq,
               step, args, dataset, epoch, tsne, cluster=True):
    
    if cluster:
        #log_to_tensorboard(summary_writer, {"Embs average": np.mean(node_new_features), 'Embs std': np.std(node_new_features) }, step)
        tsne_path = os.path.join(args.outdir, f"{args.outname}_tsne_clusters_epoch_{epoch}.png")
        cluster_labels, stats, positive_pairs, hq_bins = compute_clusters_and_stats(
            node_new_features[cluster_mask], np.array(dataset.node_names)[cluster_mask],
            dataset, clustering=args.clusteringalgo, k=args.kclusters, tsne=tsne, tsne_path=tsne_path, max_pos_pairs=None,
            amber=(args.labels is not None and "amber" in args.labels), cuda=args.cuda,
        )
    else:
        breakpoint()
        cluster_labels = np.argmax(node_new_features[cluster_mask], axis=1)
        hq, positive_clusters = compute_hq(reference_markers=dataset.ref_marker_sets,
                                           contig_genes=dataset.contig_markers,
                                           node_names=np.array(dataset.node_names)[cluster_mask],
                                           node_labels=cluster_labels)
        mq, _ = compute_hq(reference_markers=dataset.ref_marker_sets,
                                           contig_genes=dataset.contig_markers,
                                           node_names=np.array(dataset.node_names)[cluster_mask],
                                           node_labels=cluster_labels, comp_th=50, cont_th=10,)
        stats = {"hq": hq, "unresolved": len(positive_clusters), "mq": mq}
        
        cluster_to_contig = {i: [] for i in range(max(cluster_labels) + 1)}
        for i in range(len(dataset.node_names)):
            cluster_to_contig[cluster_labels[i]].append(dataset.node_names[i])
        positive_pairs = None
   
    stats["epoch"] = epoch
    #logger.info(str(stats))

    #log_to_tensorboard(summary_writer, {"hq_bins": stats["hq"], "mq_bins": stats["mq"]}, step)
    #all_cluster_labels.append(cluster_labels)
    new_best = False
    if dataset.contig_markers is not None and stats["hq"] > best_hq:
        new_best = True
    elif dataset.contig_markers is None and stats["f1"] > best_hq:
        new_best = True
    # print('--- END ---')
    #if args.quiet:
    #    logger.info(f"--- EPOCH {e:d} ---")
    #    logger.info(f"[{gname} {nlayers_gnn}l] L={gnn_loss:.3f} D={diff_loss:.3f} HQ={stats['hq']} BestHQ={best_hq} Best Epoch={best_epoch} Max GPU MB={gpu_mem_alloc:.1f}")
    #    logger.info(str(stats))

    #cluster_labels, stats, _, hq_bins = compute_clusters_and_stats(
    #    node_new_features, np.array(dataset.node_names),
    #    dataset, clustering=args.clusteringalgo, k=args.kclusters, tsne=args.tsne, #cuda=args.cuda,
    #)
    #log_to_tensorboard(summary_writer, {"hq_bins_all": stats["hq"], "mq_bins_all": stats["mq"]}, step)

    return stats, new_best, cluster_labels, positive_pairs

def main():
    sets = read_marker_gene_sets(sys.argv[1])
    markers = read_contig_genes(sys.argv[2])
    #print("Single contig bin eval")
    if len(sys.argv) < 4:
        print("Single contig bin eval")
        single_contig_bins = {c: [c] for c in markers}
        results = evaluate_contig_sets(sets, markers, single_contig_bins)
        total_hq = 0
        for bin in results:
            if results[bin]["comp"] > 90 and results[bin]["cont"] < 5:
                print(bin, results[bin]["comp"], results[bin]["cont"])
                total_hq += 1
        print("Total HQ", total_hq)
        # for bin in markers:
        #    print(bin)
        #    print(completeness(sets, set(markers[bin].keys())))
        #    print(contamination(sets, markers[bin]))
    else:
        cluster_to_contigs = {}
        clusters_file = sys.argv[3]
        with open(clusters_file, "r") as f:
            for line in f:
                if line.startswith("@") or line.startswith("#"):
                    continue
                values = line.strip().split()
                contig, binname = values[0], values[1]
                if binname not in cluster_to_contigs:
                    cluster_to_contigs[binname] = []
                cluster_to_contigs[binname].append(contig)
        results = evaluate_contig_sets(sets, markers, cluster_to_contigs)
        # print(results)

        total_hq = 0
        for bin in results:
            if results[bin]["comp"] > 90 and results[bin]["cont"] < 5:
                print(bin, results[bin]["comp"], results[bin]["cont"])
                total_hq += 1
        print("Total HQ", total_hq)
        if len(sys.argv) > 4:
            import graphmb.amber_eval as amber
            scores = {}
            amber_metrics, bin_counts = amber.amber_eval(sys.argv[4], clusters_file, ["graphmb"])
            scores["precision_avg_bp"] = amber_metrics["precision_avg_bp"]
            scores["recall_avg_bp"] = amber_metrics["recall_avg_bp"]
            scores["f1_avg_bp"] = amber_metrics["f1_avg_bp"]
            print(scores)
        # count how many contigs per marker gene
    scg_counts = {}
    for marker_set in sets:
        for gene in marker_set:
            scg_counts[gene] = 0
            for contig in markers:
                if gene in markers[contig]:
                    scg_counts[gene] += markers[contig][gene]

    #print(scg_counts)
    quartiles = np.percentile(list(scg_counts.values()), [25, 50, 75])
    print("candidate k0s", sorted(set([k for k in scg_counts.values() if k >= quartiles[2]])))

    # TODO convert SCGs to features


if __name__ == "__main__":
    main()




import time
import os
import networkx as nx
import numpy as np
import operator

#import tensorflow

from graphmb.evaluate import read_contig_genes, read_marker_gene_sets, evaluate_contig_sets
from graphmb.contigsdataset import get_kmer_to_id, count_kmers
from graphmb.utils import SEED
# import torch

def open_gfa_file(filename, filter=1000, root=False, kmer=4):
    G = nx.Graph()
    if root:
        root_node = G.add_node("root", length=0)
    skipped_contigs = set()
    skipped_edges = set()
    kmer_to_id, canonical_k = get_kmer_to_id(kmer)
    with open(filename, "r") as f:
        for line in f:
            if line.startswith("S"):
                values = line.strip().split()
                contigid = values[1]
                # contiglen = int(values[3].split(":")[-1])
                contiglen = len(values[2])
                contig_kmers = count_kmers(values[2], kmer, kmer_to_id, canonical_k)
                if contiglen < filter:
                    skipped_contigs.add(contigid)
                else:

                    G.add_node(contigid, length=contiglen, kmers=contig_kmers)  # TODO: add coverage and kmer too
                    if root:
                        G.add_edge(contigid, "root")
            if line.startswith("L"):
                values = line.strip().split()
                if values[1] in skipped_contigs or values[3] in skipped_contigs:
                    skipped_edges.add((values[1], values[3]))
                    continue
                G.add_edge(values[1], values[3])

    print("load graph from GFA")
    print("skipped these contigs {} (len<{})".format(len(skipped_contigs), filter))
    print("skipped these edges {} (len<{})".format(len(skipped_edges), filter))
    return G


def cluster_embs(node_embeddings, node_ids, clusteringalgo, kclusters, device="cpu", node_lens=None, seed=0):
    #set_seed(seed)
    if clusteringalgo == "vamb":
        from vamb.cluster import cluster as vamb_cluster
        it = vamb_cluster(
            node_embeddings, node_ids, cuda=(device == "cuda:0")
        )  # , seeds=seeds)  # contig_lens=node_lens)  #
        cluster_to_contig = {i: c for (i, (n, c)) in enumerate(it)}

        # get embs to clusters (cluster to contig has only contig names, not index)
        """cluster_to_embs = {
            c: np.array([node_embeddings[i] for i, n in enumerate(node_ids) if n in cluster_to_contig[c]])
            for c in cluster_to_contig
        }
        cluster_centroids = np.array([cluster_to_embs[c].mean(0) for c in cluster_to_contig])"""
        cluster_centroids = None  # not necessary for now
    else:
        from sklearn.cluster import (
            KMeans,
            DBSCAN,
            AgglomerativeClustering,
            MiniBatchKMeans,
            SpectralClustering,
            Birch,
            OPTICS,
        )
        from sklearn.mixture import GaussianMixture

        if clusteringalgo == "kmeans":
            clustering = KMeans(n_clusters=kclusters, random_state=SEED)
            cluster_labels = clustering.fit_predict(node_embeddings)
            cluster_centroids = clustering.cluster_centers_
        elif clusteringalgo == "kmeansgpu":
            node_embeddings = torch.tensor(node_embeddings).cuda()
            cluster_labels, cluster_centroids = kmeans_pytorch.kmeans(
                X=node_embeddings, num_clusters=kclusters, device=torch.device("cuda:0")
            )
        elif clusteringalgo == "dbscan":
            cluster_labels = DBSCAN(eps=1.1, min_samples=5).fit_predict(node_embeddings)
            # cluster_centroids = clustering.cluster_centers_
            cluster_centroids = None
        elif clusteringalgo == "gmm":
            cluster_model = GaussianMixture(
                n_components=kclusters,
                covariance_type="full",
                max_iter=1000,
                random_state=SEED,
                verbose=2,
                verbose_interval=1,
            ).fit(node_embeddings)
            cluster_labels = cluster_model.predict(node_embeddings)
            cluster_centroids = cluster_model.means_
        elif clusteringalgo == "kmeansconst":
            cluster_labels = KMeansConstrained(
                n_clusters=kclusters, size_min=1, size_max=5, random_state=SEED
            ).fit_predict(node_embeddings)
        elif clusteringalgo == "kmeansbatch":
            kmeans = MiniBatchKMeans(n_clusters=kclusters, random_state=SEED, batch_size=100, init=seed_matrix)
            cluster_labels = kmeans.fit_predict(node_embeddings)
            cluster_centroids = kmeans.cluster_centers_
        elif clusteringalgo == "spectral":
            cluster_labels = SpectralClustering(n_clusters=kclusters, random_state=SEED).fit_predict(node_embeddings)
            cluster_centroids = None
        elif clusteringalgo == "birch":
            cluster_labels = Birch(n_clusters=kclusters).fit_predict(node_embeddings)
            cluster_centroids = None
        elif clusteringalgo == "optics":
            cluster_labels = OPTICS(min_samples=5, cluster_method="xi", n_jobs=-1).fit_predict(node_embeddings)
            cluster_centroids = None
        else:
            print("invalid clustering algorithm")
            return None
        # compare clustering labels with actual labels
        # for each cluster, get majority label, and then P/R
        cluster_to_contig = {i: [] for i in range(kclusters)}
        for il, l in enumerate(cluster_labels):
            # if l not in cluster_to_contig:
            #    cluster_to_contig[l] = []
            cluster_to_contig[l].append(node_ids[il])

    return cluster_to_contig, cluster_centroids


def evaluate_binning(cluster_to_contig, node_to_label, label_to_node, outputclusters=False, contig_sizes=None):
    """Evaluate binning results using contig labels (supervised scenario)

    :param cluster_to_contig: mapping cluster IDs to contig names
    :type cluster_to_contig: dict
    :param node_to_label: mapping contig ids to labels (taxon)
    :type node_to_label: dict
    :param label_to_node: mapping label names to cotnigs
    :type label_to_node: dict
    :param outputclusters: print members of all clusters, defaults to False
    :type outputclusters: bool, optional
    :param contig_sizes: Provide contig sizes for balanced scores, defaults to None
    :type contig_sizes: dict, optional
    """
    # uncomment if you want to plot only the biggest labels
    # labels_to_plot = sorted(label_to_node, key = lambda key: len(label_to_node.get(key, [])), reverse=True)[:6]
    # print(labels_to_plot)
    avg_precision = []
    avg_recall = []
    avg_f1 = []
    cluster_sizes = []
    if contig_sizes is None:
        contig_sizes = {c: 1 for c in node_to_label}
    for c in cluster_to_contig:
        cluster_labels = [node_to_label.get(n) for n in cluster_to_contig[c] if n in node_to_label]
        cluster_counts = {}
        for label in set(cluster_labels):
            cluster_counts[label] = sum(
                [contig_sizes[n] for n in cluster_to_contig[c] if node_to_label.get(n) == label]
            )
        if len(cluster_labels) == 0:  # we do not have labels for any of the elements of this cluster
            continue
        # get majority label:
        # cluster_counter = collections.Counter(cluster_labels)
        # cluster_majority = cluster_counter.most_common(1)
        cluster_majority = max(cluster_counts.items(), key=operator.itemgetter(1))
        # print(cluster_majority)
        # if cluster_majority[0][0] not in labels_to_plot:
        #    continue
        if cluster_majority[0] not in label_to_node:
            print(
                "tie cluster",
                f"cluster {c}, majority: {cluster_majority[0]}, cluster size {len(cluster_to_contig[c])}",
            )
            continue
        # print(f"cluster {c}, majority: {cluster_majority}, cluster size {len(cluster_to_contig[c])}")
        # print(f" {len(label_to_node.get(cluster_majority[0], []))} contigs with this label")
        # avg_precision.append(len(cluster_to_contig) * cluster_majority[0][1]/len(cluster_to_contig[c]))
        # avg_recall.append(len(cluster_to_contig) * cluster_majority[0][1]/len(label_to_node[cluster_majority[0][0]]))
        cluster_size = sum([contig_sizes.get(n, 1) for n in cluster_to_contig[c]])
        cluster_p = cluster_majority[1] / cluster_size
        avg_precision.append(cluster_p)
        cluster_r = cluster_majority[1] / sum([contig_sizes.get(n, 1) for n in label_to_node[cluster_majority[0]]])
        avg_recall.append(cluster_r)
        cluster_f1 = 2 * cluster_p * cluster_r / (cluster_p + cluster_r)
        avg_f1.append(cluster_f1)
        # print(cluster_p, cluster_r)
        if outputclusters:
            print(cluster_to_contig[c])
        cluster_sizes.append(len(cluster_to_contig))
    # print("average precision", sum(avg_precision)/sum(cluster_sizes)/len(avg_precision))
    # print("average recall", sum(avg_recall)/sum(cluster_sizes)/len(avg_recall))
    print(
        "average precision",
        round(sum(avg_precision) / len(avg_precision), 4),
        "average recall",
        round(sum(avg_recall) / len(avg_recall), 4),
        "average f1",
        round(sum(avg_f1) / len(avg_f1), 4),
        "P>0.95 and R>0.9:",
        len([i for i in range(len(avg_recall)) if avg_recall[i] >= 0.9 and avg_precision[i] >= 0.95]),
    )


def read_contigs_scg(ref_file, marker_file, node_names):
    """Read marker stats and return count table

    :param ref_file: path to file with reference markers (e.g. Bacteria.ms)
    :type ref_file: str
    :param marker_file: path to file with markers found on each contig
    :type marker_file: str
    :param node_names: list of node names
    :type node_names: list
    :return: matrix where rows are contigs, columns are markers, values are marker counts
    :rtype: [type]
    """
    ref_marker_genes = read_marker_gene_sets(ref_file)
    contigs_markers = read_contig_genes(marker_file)

    # flatten ref marker gene sets
    ref_marker_genes = [g for m in ref_marker_genes for g in m]
    counts = []
    for contig in node_names:
        counts.append([contigs_markers[contig].get(g, 0) for g in ref_marker_genes])

    return np.array(counts)


def calculate_bin_metrics(results, extra_metrics=False):
    """Calculate overall scores over a set of bins for which we already have scores

    :param results: scores of each bin
    :type results: dict
    :param extra_metrics: Calculate more metrics, defaults to False
    :type extra_metrics: bool, optional
    :return: overall metrics for this bin set
    :rtype: dict
    """
    hq_bins = [bin for bin in results if results[bin]["comp"] >= 90 and results[bin]["cont"] < 5]
    mq_bins = [bin for bin in results if results[bin]["comp"] >= 50 and results[bin]["cont"] < 10]
    metrics = {"hq": hq_bins, "mq": mq_bins, "total": results}
    if extra_metrics and len(results) > 0:
        metrics["avg_comp"] = sum([results[bin]["comp"] for bin in results]) / len(results)
        metrics["avg_cont"] = sum([results[bin]["cont"] for bin in results]) / len(results)
        cont_comp50 = [results[bin]["cont"] for bin in results if results[bin]["comp"] > 50]
        metrics["cont_comp50"] = sum(cont_comp50) / len(cont_comp50) if len(cont_comp50) > 0 else 0
        cont_comp90 = [results[bin]["cont"] for bin in results if results[bin]["comp"] > 90]
        metrics["cont_comp90"] = sum(cont_comp90) / len(cont_comp90) if len(cont_comp90) > 0 else 0
        comp_cont5 = [results[bin]["comp"] for bin in results if results[bin]["cont"] < 5]
        metrics["comp_cont5"] = sum(comp_cont5) / len(comp_cont5) if len(comp_cont5) > 0 else 0
    return metrics


def cluster_eval(
    model,
    dataset,
    logits,
    clustering,
    k,
    loss,
    best_hq,
    best_hq_epoch,
    epoch,
    device,
    clusteringloss=False,
    logger=None,
    use_marker_contigs_only=False,
    seed=0,
):
    """Cluster contig embs and eval with markers

    :param model: Model used to generate embs, save if better than best_hq
    :type model: nn.Module
    :param dataset: dataset object used to train model
    :type dataset: ContigsDataset
    :param logits: tensor with output of model
    :type logits: torch.Tensor
    :param clustering: Type of clustering to be done
    :type clustering: str
    :param k: Number of clusters
    :type k: int
    :param loss: loss (for logging)
    :type loss: [type]
    :param best_hq: Best HQ obtained at this point
    :type best_hq: int
    :param best_hq_epoch: Epoch where best HQ was obtained
    :type best_hq_epoch: int
    :param epoch: Current epoch
    :type epoch: int
    :param device: If using cuda for clustering
    :type device: str
    :param clusteringloss: Compute a clustering loss, defaults to False
    :type clusteringloss: bool, optional
    :param logger: Logger object, defaults to None
    :type logger: [type], optional
    :return: new best HQ and epoch, clustering loss, cluster to contig mapping
    :rtype: list
    """
    import torch
    kmeans_loss = None
    t0_cluster = time.time()
    model.eval()
    #set_seed(seed)
    if torch.is_tensor(logits):
        embeds = logits.detach().numpy()
    else:
        embeds = logits
    if use_marker_contigs_only:
        marker_mask = [n in dataset.contig_markers and len(dataset.contig_markers[n]) > 0 for n in dataset.node_names]
        print("clustering", sum(marker_mask), "markers", len(marker_mask))
        cluster_embeds = embeds[marker_mask]
        cluster_names = np.array(dataset.node_names)[marker_mask]
    else:
        cluster_embeds = embeds
        cluster_names = dataset.node_names
    cluster_to_contig, centroids = cluster_embs(
        cluster_embeds,
        cluster_names,
        clustering,
        # len(dataset.connected),
        k,
        device=device,
        seed=seed,
    )
    contig_to_cluster = {}
    for bin in cluster_to_contig:
        for contig in cluster_to_contig[bin]:
            contig_to_cluster[contig] = bin

    if dataset.ref_marker_sets is not None:
        results = evaluate_contig_sets(dataset.ref_marker_sets, dataset.contig_markers, cluster_to_contig)
        metrics = calculate_bin_metrics(results)
        logger.info(
            f"HQ: {len(metrics['hq'])}, MQ:, {len(metrics['mq'])} Total bins: {len(metrics['total'])} Best HQ: {best_hq} Best HQ epoch: {best_hq_epoch}"
        )
        # logger.info("HQ:", hq_bins, "avg comp", avg_comp, "avg cont", avg_cont, "avg cont when comp>50", cont_comp50)
        # logger.info(
        #    "HQ {}, avg comp {:.2f}, avg cont {:.2f}, cont when comp>50 {:.2f}, cont when comp>90 {:.2f}, comp when cont<5 {:.2f}".format(
        #        hq_bins, avg_comp, avg_cont, cont_comp50, cont_comp90, comp_cont5
        #    )
        # )
        # print("clustering time:", time.time() - t0_cluster, embeds.shape)
        if len(metrics["hq"]) > best_hq:
            best_hq = len(metrics["hq"])
            best_hq_epoch = epoch
            logger.info("new best checkpoint, saving checkpoint to best_model_hq.pkl")
            torch.save(model.state_dict(), os.path.join(dataset.cache_dir, "best_model_hq.pkl"))

        if clusteringloss:
            centroids = torch.tensor(centroids, device=device)
            # get centroids of hq bins
            # centroids = centroids[hq]
            # get IDs of contigs of HQ bins
            hq_node_ids = [
                dataset.node_names.index(n)
                for bin in metrics["hq"]
                for n in cluster_to_contig[bin]
                # if sum(dataset.contig_markers[n].values()) > 0
            ]
            # count only nodes with SCGs
            # TODO multiply squared dist by completness of cluster
            # cluster_comp = np.array([results[i]["comp"] / 100 for i, r in enumerate(results)])
            # breakpoint()
            cluster_cont = np.array([i for i, r in enumerate(results) if results[i]["cont"] > 10])

            cont_node_ids = [
                dataset.node_names.index(n)
                for bin in cluster_cont
                for n in cluster_to_contig[bin]
                # if sum(dataset.contig_markers[n].values()) > 0
            ]
            # TODO subtract dist of contaminated bins (same num of good bins)
            hq_logits = logits[hq_node_ids]
            # cont_logits = logits[cont_node_ids]
            # breakpoint()
            # https://discuss.pytorch.org/t/k-means-loss-calculation/22041/7
            kmeans_loss_good = ((hq_logits[:, None] - centroids[None]) ** 2).sum(2).min(1)[0].mean()
            # kmeans_loss_bad = ((cont_logits[:, None] - centroids[None]) ** 2).sum(2).min(1)[0].mean()
            # kmeans_loss = -kmeans_loss_bad
            kmeans_loss = kmeans_loss_good
            # kmeans_loss = kmeans_loss_good - kmeans_loss_bad
            # kmeans_loss = ((logits[:, None] - centroids[None]) ** 2).sum(2).min(1)[0].mean()
            logger.info(
                f"Kmeans loss: {kmeans_loss.item()} on {len(hq_logits)} points {len(centroids)} total clusters",
            )
            # loss = kmeans_loss * alpha + (1 - alpha) * loss
    logger.info(
        "Epoch {:05d} | Best HQ: {} | Best epoch {} | Total loss {:.4f}".format(
            epoch,
            best_hq,
            best_hq_epoch,
            loss.detach(),
        )
    )
    return best_hq, best_hq_epoch, kmeans_loss, cluster_to_contig, metrics


def compute_loss_para(adj, device):
    pos_weight = (adj.shape[0] * adj.shape[0] - adj.sum()) / adj.sum()
    norm = adj.shape[0] * adj.shape[0] / float((adj.shape[0] * adj.shape[0] - adj.sum()) * 2)
    weight_mask = adj.view(-1) == 1
    weight_tensor = torch.ones(weight_mask.size(0)).to(device)
    weight_tensor[weight_mask] = pos_weight
    return weight_tensor, norm



import itertools
import argparse
import shutil
import random
import time
import logging
from datetime import datetime
from collections import Counter
import numpy as np
import sys
import os
import copy
import pickle
import shutil
import pdb
import networkx as nx

from pathlib import Path
from graphmb.arg_options import create_parser
from graphmb.evaluate import (
    evaluate_contig_sets,
    calculate_overall_prf,
)
from graphmb.contigsdataset import AssemblyDataset
from graphmb.visualize import draw_nx_graph, run_tsne, plot_embs
from graphmb.utils import set_seed, get_cluster_mask
from graphmb.graphmb1 import (cluster_embs,
    evaluate_binning,
    calculate_bin_metrics,
    )

from graphmb.version import __version__

def run_model(dataset, args, logger, nrun, target_metric):
    if args.model_name.endswith("_ae"):
        from graphmb import train_vaegnn
        return train_vaegnn.run_model_vaegnn(dataset, args, logger, nrun, target_metric)
    elif args.model_name == "vae":
        from graphmb import train_vae
        return train_vae.run_model_vae(dataset, args, logger, nrun)
    elif args.model_name in ("gcn", "sage", "gat"):
        from graphmb import train_gnn
        return train_gnn.run_model_gnn(dataset, args, logger, nrun)

def draw(dataset, node_to_label, label_to_node, cluster_to_contig, outname, graph=None):
    # properties of all nodes
    nodeid_to_label = {i: node_to_label.get(n, "NA") for i, n in enumerate(dataset.node_names)}
    contig_lens = {i: dataset.node_lengths[i] for i in range(len(dataset.node_names))}
    nodes_titles = {
        i: str(dataset.node_names[i]) + "<br>Length: " + str(contig_lens[i]) for i in range(len(dataset.contig_names))
    }
    if dataset.depth is not None:
        nodes_titles = {
            i: nodes_titles[i] + "<br>Depth: " + ", ".join(["{:.4}".format(x) for x in dataset.nodes_depths[i]])
            for i in range(len(dataset.contig_names))
        }
    if cluster_to_contig:
        contig_to_cluster = {contig: cluster for cluster, contigs in cluster_to_contig.items() for contig in contigs}
        nodes_titles = {
            i: nodes_titles[i] + "<br>Cluster: " + str(contig_to_cluster[n])
            for i, n in enumerate(dataset.contig_names)
        }

    # convert DGL graph to networkx
    nx_graph = graph.cpu().to_networkx(edge_attrs=["weight"]).to_undirected()
    connected_comp = [c for c in sorted(nx.connected_components(nx_graph), key=len, reverse=True) if len(c) > 0]
    # draw connected components to separate files
    for i in range(10, 50):
        # without_largest_comp = [item for sublist in connected_comp[10:110] for item in sublist if len(sublist) > 2]
        this_comp = connected_comp[i]
        nx_graph = nx.subgraph(nx_graph, this_comp)

        draw_nx_graph(
            nx_graph,
            nodeid_to_label,
            label_to_node,
            outname + "_" + str(i),
            contig_sizes=contig_lens,
            node_titles=nodes_titles,
        )


def write_embs(embs, node_names, outname):
    # write embs as dict node_name: embs
    embs_dict = {node_names[i]: embs[i] for i in range(len(embs))}
    with open(outname, "wb") as f:
        pickle.dump(embs_dict, f)

def write_edges(graph, outname):
    with open(outname, "w") as graphf:
        for e in zip(graph.edges()[0], graph.edges()[1]):
            graphf.write(str(e[0].item()) + "\t" + str(e[1].item()) + "\n")


def check_dirs(args, use_features=True):
    """Check if files necessary to run exist, other wise print message and exit"""
    if args.outdir is None:
        if args.assembly is None:
            print("Please specify assembly path or outdir with --assembly or --outdir")
            exit()
        else:
            args.outdir = args.assembly
    else:
        Path(args.outdir).mkdir(parents=True, exist_ok=True)

    # check if other dirs exists
    if not os.path.exists(os.path.join(args.assembly, args.graph_file)):
        print(f"Assembly Graph file {args.graph_file} not found, check --graph_file option")
    if not os.path.exists(os.path.join(args.assembly, args.features)) or not use_features:

        # needs assembly files to calculate features
        if not os.path.exists(os.path.join(args.assembly, args.assembly_name)):
            print(f"Assembly {args.assembly_name} not found, check --assembly_name option")
            exit()
        if not os.path.exists(os.path.join(args.assembly, args.depth)):
            print(f"Depth file {args.depth} not found, check --depth option, not using depths")
            #exit()


def get_activation(args):
    # pick activation function
    import torch.nn as nn
    if args.activation == "prelu":
        activation = nn.PReLU(args.hidden_gnn)
    elif args.activation == "relu":
        activation = nn.ReLU()
    elif args.activation == "tanh":
        activation = nn.Tanh()
    elif args.activation == "sigmoid":
        activation = nn.Sigmoid()
    elif args.activation == "lrelu":
        activation = nn.LeakyReLU()
    return activation


def run_graphmb(dataset, args, device, logger):
    from graphmb.graphsage_unsupervised import train_graphsage, SAGE
    activation = get_activation(args)
    model = SAGE(
        dataset.graph.ndata["feat"].shape[1],
        args.hidden_gnn,
        args.embsize_gnn,
        args.layers_gnn,
        activation,
        args.dropout_gnn,
        agg=args.aggtype
    )
    model = model.to(device)

    if model is not None:
        model = model.to(device)

    cluster_mask = get_cluster_mask(args.quick, dataset.assembly)
    logging.info(model)
    if (
        dataset.assembly.ref_marker_sets is not None
        and args.clusteringalgo is not None
        and not args.skip_preclustering
    ):
        # cluster using only input features
        logger.info("pre train clustering:")
        pre_cluster_to_contig, centroids = cluster_embs(
            dataset.graph.ndata["feat"].detach().cpu().numpy()[cluster_mask],
            np.array(dataset.assembly.node_names)[cluster_mask],
            args.clusteringalgo,
            args.kclusters,
            device=device,
            # node_lens=np.array([c[0] for c in dataset.assembly.nodes_len]),
            node_lens=np.array(dataset.assembly.node_lengths)[cluster_mask],
            seed=args.seed,
        )
        results = evaluate_contig_sets(
            dataset.assembly.ref_marker_sets, dataset.assembly.contig_markers, pre_cluster_to_contig
        )
        metrics = calculate_bin_metrics(results)
        logger.info(f"HQ: {len(metrics['hq'])}, MQ:, {len(metrics['mq'])} Total bins: {len(metrics['total'])}")

    best_train_embs, best_model, last_train_embs, last_model, metrics = train_graphsage(
        dataset,
        model,
        batch_size=args.batchsize,
        fan_out=args.fanout,
        num_negs=args.negatives,
        neg_share=False,
        num_epochs=args.epoch,
        lr=args.lr_gnn,
        k=args.kclusters,
        clusteringalgo=args.clusteringalgo,
        cluster_features=args.concatfeatures,
        print_interval=args.print,
        loss_weights=(not args.no_loss_weights),
        sample_weights=(not args.no_sample_weights),
        logger=logger,
        device=device,
        epsilon=args.early_stopping,
        evalepochs=args.evalepochs,
        seed=args.seed,
        eval_skip=args.evalskip,
        quick=args.quick
    )
    return best_train_embs, best_model, last_train_embs, last_model, metrics


def write_bins(args, dataset, cluster_to_contig, logger):
    bin_dir = Path(args.outdir + "/{}_bins/".format(args.outname))
    bin_dir.mkdir(parents=True, exist_ok=True)
    [f.unlink() for f in bin_dir.glob("*.fa") if f.is_file()]
    clustered_contigs = set()
    multi_contig_clusters = 0
    logger.info(f"{len(cluster_to_contig)} clusters")
    short_contigs = set()
    skipped_clusters = 0
    for c in cluster_to_contig:
        cluster_size = sum([dataset.node_lengths[dataset.node_names.index(contig)] for contig in cluster_to_contig[c]])
        if cluster_size < args.minbin:
            # print("skipped small cluster", len(cluster_to_contig[c]), "contig")
            for contig in cluster_to_contig[c]:
                short_contigs.add(contig)
            skipped_clusters += 1
            continue
        multi_contig_clusters += 1
        with open(bin_dir / f"{c}.fa", "w") as binfile:
            for contig in cluster_to_contig[c]:
                binfile.write(">" + contig + "\n")
                binfile.write(dataset.node_seqs[contig] + "\n")
                clustered_contigs.add(contig)
        # print("multi cluster", c, "size", cluster_size, "contigs", len(cluster_to_contig[c]))
    logger.info("### skipped {} clusters while writing to file".format(skipped_clusters))
    single_clusters = multi_contig_clusters
    left_over = set(dataset.node_names) - clustered_contigs - short_contigs
    for c in left_over:
        if c not in clustered_contigs and len(dataset.node_seqs[c]) > args.minbin:

            with open(bin_dir / f"{single_clusters}.fna", "w") as binfile:
                binfile.write(">" + c + "\n")
                binfile.write(dataset.node_seqs[c] + "\n")
                single_clusters += 1
            # print("contig", single_clusters, "size", len(dataset.contig_seqs[c]))
    logger.info(f"### wrote {single_clusters} clusters {multi_contig_clusters} >= #contig {args.mincomp}")


def run_post_processing(final_embs, args, logger, dataset, device, label_to_node, node_to_label, seed):
    metrics = {}
    if "cluster" in args.post or "kmeans" in args.post:
        logger.info("#### clustering embs with {} ({})".format(args.clusteringalgo, args.kclusters))
        # train_embs = last_train_embs

        if args.clusteringalgo is False:
            args.clusteringalgo = "kmeans"
        if not isinstance(final_embs, np.ndarray):
            if args.cuda:
                final_embs = final_embs.cpu()
            final_embs = final_embs.numpy()

            # last_train_embs should already be detached and on cpu
        best_cluster_to_contig, best_centroids = cluster_embs(
            final_embs,
            dataset.node_names,
            args.clusteringalgo,
            # len(dataset.connected),
            args.kclusters,
            device=device,
            seed=seed,
        )
        cluster_sizes = {}
        for c in best_cluster_to_contig:
            cluster_size = sum([len(dataset.node_seqs[contig]) for contig in best_cluster_to_contig[c]])
            cluster_sizes[c] = cluster_size
        best_contig_to_bin = {}
        for bin in best_cluster_to_contig:
            for contig in best_cluster_to_contig[bin]:
                best_contig_to_bin[contig] = bin
        # run for best epoch only
        if args.markers is not None:
            total_hq = 0
            total_mq = 0
            results = evaluate_contig_sets(dataset.ref_marker_sets, dataset.contig_markers, best_cluster_to_contig)
            hq_bins = set()
            for binid in results:
                if results[binid]["comp"] > 90 and results[binid]["cont"] < 5:
                    contig_labels = [dataset.node_to_label.get(node, 0) for node in best_cluster_to_contig[binid]]
                    labels_count = Counter(contig_labels)
                    # logger.debug(
                    #    f"{binid}, {round(results[binid]['comp'],4)}, {round(results[binid]['cont'],4)}, "
                    #    f"{len(best_cluster_to_contig[binid])} {labels_count}"
                    # )
                    hq_bins.add(binid)
                    total_hq += 1
                if results[binid]["comp"] > 50 and results[binid]["cont"] < 10:
                    total_mq += 1
            logger.info("#### Total HQ {} ####".format(total_hq))
            logger.info("#### Total MQ {} ####".format(total_mq))
            metrics["hq_bins"] = total_hq
            metrics["mq_bins"] = total_mq
        contig_lens = {dataset.node_names[i]: dataset.node_lengths[i] for i in range(len(dataset.node_names))}
        if len(dataset.labels) > 1:
            evaluate_binning(best_cluster_to_contig, node_to_label, label_to_node, contig_sizes=contig_lens)
            # calculate overall P/R/F
            calculate_overall_prf(best_cluster_to_contig, best_contig_to_bin, node_to_label, label_to_node, contig_lens)
            calculate_overall_prf(
                {
                    cluster: best_cluster_to_contig[cluster]
                    for cluster in best_cluster_to_contig
                    if cluster_sizes[cluster] > args.minbin
                },
                {
                    contig: best_contig_to_bin[contig]
                    for contig in best_contig_to_bin
                    if cluster_sizes[best_contig_to_bin[contig]] > args.minbin
                },
                node_to_label,
                label_to_node,
                contig_lens
            )
        if "writebins" in args.post:
            logger.info(f"### writing bins to {args.outdir}/{args.outname}_bins/")
            write_bins(args, dataset, best_cluster_to_contig, logger)
        if "contig2bin" in args.post:
            # invert cluster_to_contig
            logger.info("### Writing contig2bin to {}/{}".format(args.outdir, args.outname))
            with open(args.outdir + f"/{args.outname}_best_contig2bin.tsv", "w") as f:
                f.write("@Version:0.9.0\n@SampleID:SAMPLEID\n@@SEQUENCEID\tBINID\n")
                for c in best_contig_to_bin:
                    f.write(f"{str(c)}\t{str(best_contig_to_bin[c])}\n")

    # plot tsne embs
    if "tsne" in args.post:
        node_embeddings_2dim, centroids_2dim = run_tsne(final_embs, dataset, best_cluster_to_contig, hq_bins)
        plot_embs(
            dataset.node_names,
            node_embeddings_2dim,
            label_to_node,
            centroids=centroids_2dim,
            hq_centroids=hq_bins,
            node_sizes=None,
            outputname=os.path.join(args.outdir, args.outname + "_tsne_clusters.png"),
        )

        # node_sizes=[dataset.nodes_len[i][0] * 100 for i in range(len(dataset.contig_names))],
    if "draw" in args.post:
        print("drawing graph")
        draw(
            dataset,
            node_to_label,
            label_to_node,
            best_cluster_to_contig,
            os.path.join(args.outdir, args.outname + "_graph.png"),
            graph=graph,
        )

    if "edges" in args.post:
        logger.info(f"### writing edges to {args.outdir + args.outname}_edges")
        write_edges(graph, os.path.join(args.outdir, args.outname + "_edges"))

    if "writeembs" in args.post:
        logger.info("### writing best and last embs to {}".format(args.outdir))
        write_embs(final_embs, dataset.node_names, os.path.join(args.outdir, f"{args.outname}_best_embs.pickle"))
        # write_embs(best_train_embs, dataset.node_names, os.path.join(args.outdir, f"{args.outname}_last_embs.pickle"))
    return metrics

def main():
    parser = create_parser()
    args = parser.parse_args()

    if args.version:
        print(f"GraphMB {__version__}")
        exit(0)
    if not args.read_cache:
        check_dirs(args)
    # set up logging
    now = datetime.now()
    logger = logging.getLogger(__name__)
    loglevel = getattr(logging, args.loglevel.upper())
    logger.setLevel(loglevel)
    logfile = os.path.join(args.outdir, now.strftime("%Y%m%d-%H%M%S") + "{}_output.log".format(args.outname))
    output_file_handler = logging.FileHandler(logfile)
    print("logging to {}".format(logfile))
    stdout_handler = logging.StreamHandler(sys.stdout)
    logger.addHandler(output_file_handler)
    logger.addHandler(stdout_handler)
    logging.getLogger('matplotlib.font_manager').setLevel(logging.ERROR)
    logging.getLogger("matplotlib").disabled = True
    logging.getLogger("matplotlib.pyplot").disabled = True

    def handle_exception(exc_type, exc_value, exc_traceback):
        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return

        logger.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))

    sys.excepthook = handle_exception

    logger.info(f"Running GraphMB {__version__}")
    logger.debug(args)
    # setup cuda and cpu
    logger.info("using cuda: {}".format(str(args.cuda)))
    device = "cuda:0" if args.cuda else "cpu"
    clustering_device = "cuda:0" if args.cuda else "cpu"
    logger.info("setting seed to {}".format(args.seed))
    set_seed(args.seed)
    use_graph = os.path.exists(os.path.join(args.assembly, args.graph_file)) or args.read_cache
    # specify data properties for caching
    if args.features is None:
        if args.assembly != "":
            features_path = os.path.join(args.assembly, "features.tsv")
        else:
            features_path = os.path.join(args.outdir, "features.tsv")
    else:
        features_path = os.path.join(args.assembly, args.features)

    # create assembly object
    dataset = AssemblyDataset(
        name=args.outname,
        logger=logger,
        data_dir=args.assembly, #every file is appended to this this
        fastafile=args.assembly_name,
        graphfile=args.graph_file,
        depthfile=args.depth,
        scgfile=args.markers,
        labelsfile=args.labels,
        featuresfile=features_path,
        cache_dir=args.outdir,
        min_contig_length=args.mincontig,
        contignodes=args.contignodes
    )
    if args.read_cache or (dataset.check_cache(use_graph) and not args.reload):
        logger.info("Reading cache from".format(args.outdir))
        dataset.read_cache(use_graph)
    else:
        check_dirs(args, use_features=False)
        logger.info("Cache not found on {}".format(args.outdir))
        dataset.read_assembly()
        logger.info("")
    
    if args.markers.startswith("gtdb"):
        dataset.read_gtdbtk_files()
    elif args.markers != "":
        dataset.read_scgs()
    else:
        args.markers = None
    
    if os.path.exists(os.path.join(args.assembly, "assembly_info.txt")):
        logger.info("Reading assembly info file")
        dataset.read_assembly_info()
        dataset.print_circular_contigs()
    # k can be user defined or dependent on the dataset
    if args.kclusters is None:
        args.kclusters = len(dataset.labels)
    args.kclusters = int(args.kclusters)

    

    # filter graph by components
    # dataset.connected = [c for c in dataset.connected if len(c) >= args.mincomp]

    # zscore Kmer features (kmer are already loaded from reading the dataset)
    # dataset.nodes_kmer = torch.FloatTensor(stats.zscore(dataset.nodes_kmer, axis=0))

    #vamb_emb_exists = os.path.exists(features_path)
    #if not args.rawfeatures and (args.vamb or not vamb_emb_exists) and args.model_name != "vae":
    #    vamb_outdir = os.path.join(args.outdir, "vamb_out{}/".format(args.vambdim))
    #    dataset.run_vamb(vamb_outdir, args.cuda, args.vambdim)

    # reload labels from file anyway
    if args.labels is not None:
        dataset.read_labels()
    
    if args.labelgraph:
        dataset.generate_edges_based_on_labels()
        dataset.calculate_homophily()
    
    dataset.print_stats()

    # graph transformations
    # Filter edges according to weight (could be from read overlap count or depth sim)
    if not args.rawfeatures and args.model_name != "vae":
        dataset.read_features()
    metrics_per_run = []
    amber_metrics_per_run = []
    for n in range(args.nruns):
        logger.info("RUN {}".format(n))
        if args.embs is not None:  # no training, just run post processing
            emb_file = args.embs
            with open(emb_file, "rb") as embsf:
                best_embs_dict = pickle.load(embsf)
                best_train_embs = np.array([best_embs_dict[i] for i in dataset.node_names])

        # DGL specific code
        elif args.model_name == "sage_lstm":
            import torch
            from graphmb.dgl_dataset import DGLAssemblyDataset
            torch.set_num_threads(args.numcores)
            dgl_dataset = DGLAssemblyDataset(dataset)
            # initialize empty features vector
            nodes_data = torch.FloatTensor(len(dataset.node_names), 0)
            if args.rawfeatures:
                nodes_data = torch.cat(
                    (nodes_data, torch.FloatTensor(dataset.node_kmers), torch.FloatTensor(dataset.node_depths)), dim=1
                )
            # if args.depth is not None:
            #    dataset.nodes_data = torch.cat((dataset.nodes_data, dataset.nodes_depths), dim=1)
            elif args.features is not None:  # append embs
                node_embs = torch.FloatTensor(dataset.node_embs)
                nodes_data = torch.cat((nodes_data, node_embs), dim=1)
            dgl_dataset.graph.ndata["feat"] = nodes_data
            # dataset.graph.ndata["len"] = torch.Tensor(dataset.nodes_len)

            # All nodes have a self loop
            # dataset.graph = dgl.remove_self_loop(dataset.graph)
            # diff_edges = len(dataset.graph.edata["weight"])
            # dataset.graph = dgl.add_self_loop(dataset.graph)

            # max_weight = dataset.graph.edata["weight"].max().item()
            # dataset.graph.edata["weight"][diff_edges:] = max_weight
            dgl_dataset.graph.edata["weight"] = dgl_dataset.graph.edata["weight"].float()
            graph = dgl_dataset[0]
            logger.info(graph)
            graph = graph.to(device)

            model = None
            if args.embs is None and args.read_embs is False:
                best_train_embs, model, last_train_embs, last_model, metrics = run_graphmb(dgl_dataset, args, device, logger)
                emb_file = args.outdir + f"/{args.outname}_train_embs.pickle"
                metrics = {k: len(v) for k, v in metrics.items()}
            if model is None:
                best_train_embs = graph.ndata["feat"]
                last_train_embs = graph.ndata["feat"]
        
        elif args.model_name in ("sage", "gcn", "gat", "vae", "vgae") or args.model_name.endswith("_ae") or \
             args.model_name.endswith("_decode") or args.model_name.endswith("_aug"):
            if "torch" in sys.modules:
                sys.modules.pop('torch')
            os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"  # FATAL
            import tensorflow as tf
            #tf.get_logger().setLevel(logging.INFO)
            clustering_device = "cpu" # avoid tf vs torch issues
            logging.getLogger("tensorflow").setLevel(logging.INFO)
            if not args.cuda:
                tf.config.set_visible_devices([], "GPU")
            # load precomputed contigs with same SCGs (diff genomes)
            if os.path.exists(f"{dataset.cache_dir}/all_different.npy"):
                dataset.neg_pairs_idx = np.load(f"{dataset.cache_dir}/all_different.npy")
            elif args.markers is not None:
                dataset.get_all_different_idx()
                np.save(f"{dataset.cache_dir}/all_different.npy", dataset.neg_pairs_idx)
            target_metric = "f1"
            if args.markers is not None:
                target_metric = "hq"
            elif "amber" in args.labels:
                target_metric = "f1_avg_bp"
            best_train_embs, metrics = run_model(dataset, args, logger, nrun=n, target_metric=target_metric)
            tf.keras.backend.clear_session()

        run_post_processing(
            best_train_embs,
            args,
            logger,
            dataset,
            clustering_device,
            dataset.label_to_node,
            dataset.node_to_label,
            seed=args.seed,
        )

        if args.labels is not None: # or "contig2bin" in args.post:
            from graphmb.amber_eval import amber_eval
            amber_metrics, bin_counts = amber_eval(
                os.path.join(args.assembly, args.labels), args.outdir + f"/{args.outname}_{n}_best_contig2bin.tsv", ["graphmb"]
            )
        #if args.labels is not None:
            hq = bin_counts["> 90% completeness"][1]
            mq = bin_counts["> 50% completeness"][1]
            amber_metrics["hq"] = hq
            amber_metrics["mq"] = mq
            amber_metrics_per_run.append(amber_metrics)
        metrics_per_run.append(metrics)
        args.seed += 1
        set_seed(args.seed)

    metrics_names = metrics_per_run[0].keys()
    for mname in metrics_names:
        values = [m.get(mname, 0) for m in metrics_per_run]
        logger.info("### {}: {:.3f} {:.3f}".format(mname, np.mean(values), np.std(values)))
    hqs = [m["hq"] for m in metrics_per_run]
    mqs = [m["mq"] for m in metrics_per_run]
    logger.info("{:.1f} {:.1f} {:.1f} {:.1f}".format(np.mean(hqs), np.std(hqs), np.mean(mqs), np.std(mqs)))
    if args.labels is not None:
        #amber_metrics_names = amber_metrics_per_run[0].keys()
        amber_metrics_names = ["precision_avg_bp", "recall_avg_bp", "f1_avg_bp", "hq", "mq"]
        for mname in amber_metrics_names:
            values = [m[mname] for m in amber_metrics_per_run]
            logger.info("### amber eval {}: {:.4f} {:.4f} ###".format(mname, np.mean(values), np.std(values)))
    total_time = datetime.now() - now
    print("Total run time: {}".format(total_time))
    print("Seconds per run: {:.2f}".format(total_time.total_seconds() / args.nruns))

if __name__ == "__main__":
    main()



from pathlib import Path
import time
import os
import sys
import math
import pdb
import itertools
from collections import Counter
import networkx as nx
import numpy as np
from tqdm import tqdm
import datetime
import operator
import scipy
import logging
# import dgl
import random
from sklearn.cluster import KMeans
#import tensorflow as tf

SEED = 0

def set_seed(seed=0):
    if "dgl" in sys.modules:
        import dgl
        print("setting dgl seed")
        dgl.random.seed(seed)
    if "torch" in sys.modules:
        import torch
        print("setting torch seed")
        torch.manual_seed(seed)
    random.seed(seed)
    np.random.seed(seed)
    if "tensorflow" in sys.modules:
        import tensorflow
        print("setting tf seed")
        tensorflow.random.set_seed(seed)



class Read:
    def __init__(self, readid, species=None):
        self.readid = readid
        self.species = species
        self.mappings = set()


class ReadMapping:
    def __init__(self, readid, bitflag, contigname, pos, mapq, seq):
        self.readid = readid
        self.bitflag = bitflag
        self.contigname = contigname
        self.pos = pos
        self.mapq = mapq


def get_cluster_mask(quick, dataset):
    if quick and dataset.contig_markers is not None:
        #connected_marker_nodes = filter_disconnected(dataset.adj_matrix, dataset.node_names, dataset.contig_markers)
        nodes_with_markers = [
            i
            for i, n in enumerate(dataset.node_names)
            if n in dataset.contig_markers and len(dataset.contig_markers[n]) > 0
        ]
        print("eval cluster with ", len(nodes_with_markers), "contigds with markers")
        cluster_mask = [n in nodes_with_markers for n in range(len(dataset.node_names))]
    else:
        cluster_mask = [True] * len(dataset.node_names)
    return cluster_mask

def save_model(args, epoch, th, th_vae):
    if th_vae is not None:
        # save encoder and decoder
        th_vae.encoder.save(os.path.join(args.outdir, args.outname + "_best_encoder"))
        th_vae.decoder.save(os.path.join(args.outdir, args.outname + "_best_decoder"))
    if th is not None:
        th.gnn_model.save(os.path.join(args.outdir, args.outname + "_best_gnn"))


def run_clustering(X, node_names, clustering_algo, cuda, k=0, tsne=False):
    
    if clustering_algo == "vamb":
        from graphmb.vamb_clustering import cluster as vamb_cluster
        starttime = datetime.datetime.now()
        X = X.astype(np.float32)
        cluster_to_contig = {
            i: c for (i, (n, c)) in enumerate(vamb_cluster(X, node_names, cuda=cuda))
        }
        clustering_time = datetime.datetime.now()
        #print("clustering time", clustering_time-starttime)
        contig_to_bin = {}
        #for b in cluster_to_contig:
        #    for contig in cluster_to_contig[b]:
        #        contig_to_bin[contig] = b
        for k, v in cluster_to_contig.items():
            contig_to_bin.update({n: k for n in v})
        labels = np.array([contig_to_bin[n] for n in node_names])
        # very slow code:
        cluster_centroids = None
        if tsne:
            cluster_to_embs = {
                c: np.array([X[i] for i, n in enumerate(node_names) if n in cluster_to_contig[c]])
                for c in cluster_to_contig
            }
            cluster_centroids = np.array([cluster_to_embs[c].mean(0) for c in cluster_to_contig])
        processing_time = datetime.datetime.now()
        #print("processing time",  processing_time - clustering_time)
    elif clustering_algo == "kmeansbatch":
        kmeans = MiniBatchKMeans(n_clusters=k, random_state=0, batch_size=2048, verbose=0) #, init=seed_matrix)
        labels = kmeans.fit_predict(X)
        contig_to_bin = {node_names[i]: labels[i] for i in range(len(node_names))}
        cluster_to_contig = {i: [] for i in range(k)}
        for i in range(len(node_names)):
            cluster_to_contig[labels[i]].append(node_names[i])
        #cluster_centroids = kmeans.cluster_centers_
    elif clustering_algo == "kmeansgpu":
        pass
    elif clustering_algo == "kmedoids":
        import kmedoids
        breakpoint()
        # TODO do this on gpu if avail
        D = np.sum((X[:,None]-X[None])**2, axis=-1)
        # TODO find best k
        km = kmedoids.KMedoids(20, method='fasterpam')
        cluster_labels = km.fit_predict(D).astype(np.int64)
    elif clustering_algo == "kmeans":
        clf = KMeans(k, random_state=1234)
        labels = clf.fit_predict(X)
        contig_to_bin = {node_names[i]: labels[i] for i in range(len(node_names))}
        cluster_to_contig = {i: [] for i in range(k)}
        for i in range(len(node_names)):
            cluster_to_contig[labels[i]].append(node_names[i])
        cluster_centroids = None
    return cluster_to_contig, contig_to_bin, labels, cluster_centroids

def filter_disconnected(adj, node_names, markers):
    # get idx of nodes that are connected or have at least one marker
    graph = nx.convert_matrix.from_scipy_sparse_matrix(adj, edge_attribute="weight")
    # breakpoint()
    nodes_to_remove = set()
    for n1 in graph.nodes:
        if len(list(graph.neighbors(n1))) == 0 or (
            node_names[n1] not in markers or len(markers[node_names[n1]]) == 0
        ):
            nodes_to_remove.add(n1)

    graph.remove_nodes_from(list(nodes_to_remove))
    assert len(graph.nodes()) == (len(node_names)-len(nodes_to_remove))
    print(f"{len(nodes_to_remove)} nodes without edges nor markers, keeping {len(graph.nodes())} nodes")
    return set(graph.nodes())

def run_model_vgae(dataset, args, logger, nrun):
    node_names = np.array(dataset.node_names)
    RESULT_EVERY = args.evalepochs
    hidden_gnn = args.hidden_gnn
    hidden_vae = args.hidden_vae
    output_dim_gnn = args.embsize_gnn
    output_dim_vae = args.embsize_vae
    epochs = args.epoch
    lr_vae = args.lr_vae
    lr_gnn = args.lr_gnn
    nlayers_gnn = args.layers_gnn
    clustering = args.clusteringalgo
    k = args.kclusters
    use_edge_weights = True
    cluster_markers_only = args.quick
    decay = 0.5 ** (2.0 / epochs)
    concat_features = args.concat_features

    current_time = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    train_log_dir = os.path.join(args.outdir, 'logs/' + args.outname + current_time + '/train')
    summary_writer = tf.summary.create_file_writer(train_log_dir)
    print("logging to tensorboard")
    tb_handler = TensorboardLogger(summary_writer, runname=args.outname + current_time)
    logger.addHandler(tb_handler)
    #tf.summary.trace_on(graph=True)

    logger.info("******* Running model: VGAE **********")
    logger.info("***** using edge weights: {} ******".format(use_edge_weights))
    logger.info("***** concat features: {} *****".format(concat_features))
    logger.info("***** cluster markers only: {} *****".format(cluster_markers_only))
    logger.info("***** threshold adj matrix: {} *****".format(args.binarize))
    logger.info("***** self edges only: {} *****".format(args.noedges))
    logger.info("***** Using raw kmer+abund features: {}".format(args.rawfeatures))
    tf.config.experimental_run_functions_eagerly(True)


    X, adj, cluster_mask, neg_pair_idx, pos_pair_idx = prepare_data_for_gnn(
            dataset, use_edge_weights, cluster_markers_only, use_raw=True,
            binarize=args.binarize, remove_edges=args.noedges)
    logger.info("***** SCG neg pairs: {}".format(neg_pair_idx.shape))
    logger.info("***** input features dimension: {}".format(X[cluster_mask].shape))
    # pre train clustering
    if not args.skip_preclustering:
        cluster_labels, stats, _, hq_bins = compute_clusters_and_stats(
                    X[cluster_mask], node_names[cluster_mask],
                    dataset, clustering=clustering, k=k, tsne=args.tsne,
                    amber=(args.labels is not None and "amber" in args.labels),
                    unresolved=True, cuda=args.cuda, 
                )
        logger.info(f">>> Pre train stats: {str(stats)}")


    model = VGAE(X.shape, hidden_dim1=hidden_gnn, hidden_dim2=output_dim_gnn, dropout=0.1,
             l2_reg=1e-5, embeddings=X, freeze_embeddings=True, lr=lr_gnn)
    X_train = np.arange(len(X))[:,None].astype(np.int64)
    A_train = tf.sparse.to_dense(adj)
    labels = dataset.adj_matrix.toarray()
    pos_weight = (adj.shape[0] * adj.shape[0] - tf.sparse.reduce_sum(adj)) / tf.sparse.reduce_sum(adj)

    norm = adj.shape[0] * adj.shape[0] / ((adj.shape[0] * adj.shape[0] - tf.sparse.reduce_sum(adj)) * 2)

    pbar_epoch = tqdm(range(epochs), disable=args.quiet, position=0)
    decay = 0.5**(2./10000)
    scores = []
    best_hq = 0
    batch_size = args.batchsize
    if batch_size == 0:
        batch_size = adj.shape[0]
    train_idx = list(range(adj.shape[0]))
    for e in pbar_epoch:
        np.random.shuffle(train_idx)
        n_batches = len(train_idx)//batch_size
        pbar_vaebatch = tqdm(range(n_batches), disable=(args.quiet or batch_size == len(train_idx) or n_batches < 100), position=1, ascii=' =')
        loss = 0
        for b in pbar_vaebatch:
            batch_idx = train_idx[b*batch_size:(b+1)*batch_size]
            loss += model.train_step(X_train, A_train, labels, pos_weight, norm, batch_idx)
        pbar_epoch.set_description(f'{loss:.3f}')
        model.optimizer.learning_rate = model.optimizer.learning_rate*decay
        gpu_mem_alloc = tf.config.experimental.get_memory_usage('GPU:0') / 1000000 if args.cuda else 0
        if (e + 1) % RESULT_EVERY == 0: # and e >= int(epochs/2):
            _, embs, _, _, _ = model((X_train, A_train), training=False)
            node_new_features = embs.numpy()

            best_hq, best_embs, best_epoch, scores, cluster_labels = eval_epoch(logger, summary_writer, node_new_features,
                                                                cluster_mask, e, args, dataset, e, scores,
                                                                best_hq, best_embs, best_epoch)

            if args.quiet:
                logger.info(f"--- EPOCH {e:d} ---")
                logger.info(f"[VGAE {nlayers_gnn}l] L={loss:.3f}  HQ={stats['hq']}  BestHQ={best_hq} Best Epoch={best_epoch} Max GPU MB={gpu_mem_alloc:.1f}")
                logger.info(str(stats))


    _, embs, _, _, _ = model((X_train, A_train), training=False)
    embs = embs.numpy()





from collections import OrderedDict
from collections import defaultdict
import itertools
import pandas as pd
from graphmb.contigsdataset import process_node_name


def write_amber_bins(contig_to_bin, outputfile):
    with open(outputfile, "w") as f:
        f.write("#\n@Version:0.9.0\n@SampleID:SAMPLEID\n@@SEQUENCEID\tBINID\n")
        for c in contig_to_bin:
            f.write(f"{str(c)}\t{str(contig_to_bin[c])}\n")

# adapt eval scripts from amber but simplified


def load_binnings(file_path_query, assemblytype, columns=["SEQUENCEID", "BINID"]):
    # columns = ["SEQUENCEID", "BINID", "TAXID", "LENGTH", "_LENGTH"]
    # sample_id_to_query_df = OrderedDict()
    #for metadata in columns:
        # logging.getLogger('amber').info('Loading ' + metadata[2]['SAMPLEID'])
        # nrows = metadata[1] - metadata[0] + 1
        # col_indices = [k for k, v in metadata[3].items() if v in columns]
        # amber files start with 4 header lines
    df = pd.read_csv(file_path_query, sep="\t", comment="#", skiprows=3, header=0)  # , usecols=col_indices)
    df = df.rename(columns={df.columns[i]: c for i, c in enumerate(columns)}, inplace=False)
    df = df.astype({'SEQUENCEID':'string'})
    df["SEQUENCEID"]= df["SEQUENCEID"].apply(process_node_name, assembly_type=assemblytype)
    if "_LENGTH" in columns:
        df.rename(columns={"_LENGTH": "LENGTH"}, inplace=True)
        df["LENGTH"] = pd.to_numeric(df["LENGTH"])
    return df


def amber_eval(gs_path, bin_path, labels=["graphmb"], assemblytype="flye"):
    gs_df = load_binnings(gs_path, columns=["SEQUENCEID", "BINID", "_LENGTH"], assemblytype=assemblytype)
    bin_df = load_binnings(bin_path, assemblytype=assemblytype)
    # load_queries(gs_df, bin_df, labels=labels, options, options_gs)
    gs_df = gs_df[["SEQUENCEID", "BINID", "LENGTH"]].rename(columns={"LENGTH": "seq_length",
                                                                       "BINID": "genome_id",
                                                                       "SEQUENCEID": "SEQUENCEID"})
    bin_df = bin_df[["SEQUENCEID", "BINID"]].rename(columns={"SEQUENCEID": "SEQUENCEID"})
    gs_df["seq_length"] = pd.to_numeric(gs_df["seq_length"])
    query_df = bin_df[["SEQUENCEID", "BINID"]]
    query_w_length = pd.merge(query_df, gs_df.drop_duplicates("SEQUENCEID"), on="SEQUENCEID", sort=False)
    query_w_length_no_dups = query_w_length.drop_duplicates("SEQUENCEID")
    gs_df_no_dups = gs_df.drop_duplicates("SEQUENCEID")
    percentage_of_assigned_bps = query_w_length_no_dups["seq_length"].sum() / gs_df_no_dups["seq_length"].sum()
    percentage_of_assigned_seqs = query_w_length_no_dups.shape[0] / gs_df_no_dups["SEQUENCEID"].shape[0]
    query_w_length_mult_seqs = query_df.reset_index().merge(gs_df, on="SEQUENCEID", sort=False)
    """if query_w_length.shape[0] < query_w_length_mult_seqs.shape[0]:
        query_w_length_mult_seqs.drop_duplicates(['index', 'genome_id'], inplace=True)
        confusion_df = query_w_length_mult_seqs.groupby(['BINID', 'genome_id'], sort=False).agg({'seq_length': 'sum', 'SEQUENCEID': 'count'}).rename(columns={'seq_length': 'genome_length', 'SEQUENCEID': 'genome_seq_counts'})

        most_abundant_genome_df = confusion_df.loc[confusion_df.groupby('BINID', sort=False)['genome_length'].idxmax()]
        most_abundant_genome_df = most_abundant_genome_df.reset_index()[['BINID', 'genome_id']]

        matching_genomes_df = pd.merge(query_w_length_mult_seqs, most_abundant_genome_df, on=['BINID', 'genome_id']).set_index('index')
        query_w_length_mult_seqs.set_index('index', inplace=True)
        difference_df = query_w_length_mult_seqs.drop(matching_genomes_df.index).groupby(['index'], sort=False).first()
        query_w_length = pd.concat([matching_genomes_df, difference_df])

        # Modify gs such that multiple binnings of the same sequence are not required
        matching_genomes_df = pd.merge(gs_df, query_w_length[['SEQUENCEID', 'genome_id']], on=['SEQUENCEID', 'genome_id'])
        matching_genomes_df = matching_genomes_df[['SEQUENCEID', 'genome_id', 'seq_length']].drop_duplicates(['SEQUENCEID', 'genome_id'])
        condition = gs_df_no_dups['SEQUENCEID'].isin(matching_genomes_df['SEQUENCEID'])
        difference_df = gs_df_no_dups[~condition]
        gs_df = pd.concat([difference_df, matching_genomes_df])"""

    # query_w_length_mult_seqs.reset_index(inplace=True)
    # query_w_length_mult_seqs = pd.merge(query_w_length_mult_seqs, most_abundant_genome_df, on=['BINID'])
    # grouped = query_w_length_mult_seqs.groupby(['index'], sort=False, as_index=False)
    # query_w_length = grouped.apply(lambda x: x[x['genome_id_x'] == x['genome_id_y'] if any(x['genome_id_x'] == x['genome_id_y']) else len(x) * [True]])
    # query_w_length = query_w_length.groupby(['index'], sort=False).first().drop(columns='genome_id_y').rename(columns={'genome_id_x': 'genome_id'})

    df = query_w_length

    confusion_df = (
        query_w_length.groupby(["BINID", "genome_id"], sort=False)
        .agg({"seq_length": "sum", "SEQUENCEID": "count"})
        .rename(columns={"seq_length": "genome_length", "SEQUENCEID": "genome_seq_counts"})
    )
    # self.confusion_df = confusion_df

    # rand_index_bp, adjusted_rand_index_bp = Metrics.compute_rand_index(
    #    confusion_df, "BINID", "genome_id", "genome_length"
    # )
    # rand_index_seq, adjusted_rand_index_seq = Metrics.compute_rand_index(
    #    confusion_df, "BINID", "genome_id", "genome_seq_counts"
    # )

    most_abundant_genome_df = (
        confusion_df.loc[confusion_df.groupby("BINID", sort=False)["genome_length"].idxmax()]
        .reset_index()
        .set_index("BINID")
    )

    query_w_length["seq_length_mean"] = query_w_length["seq_length"]

    precision_df = (
        query_w_length.groupby("BINID", sort=False)
        .agg({"seq_length": "sum", "seq_length_mean": "mean", "SEQUENCEID": "count"})
        .rename(columns={"seq_length": "total_length", "SEQUENCEID": "total_seq_counts"})
    )
    precision_df = pd.merge(precision_df, most_abundant_genome_df, on="BINID")
    precision_df.rename(columns={"genome_length": "tp_length", "genome_seq_counts": "tp_seq_counts"}, inplace=True)
    precision_df["precision_bp"] = precision_df["tp_length"] / precision_df["total_length"]
    precision_df["precision_seq"] = precision_df["tp_seq_counts"] / precision_df["total_seq_counts"]

    """if self.options.filter_tail_percentage:
        precision_df['total_length_pct'] = precision_df['total_length'] / precision_df['total_length'].sum()
        precision_df.sort_values(by='total_length', inplace=True)
        precision_df['cumsum_length_pct'] = precision_df['total_length_pct'].cumsum(axis=0)
        precision_df['precision_bp'].mask(precision_df['cumsum_length_pct'] <= self.options.filter_tail_percentage / 100, inplace=True)
        precision_df['precision_seq'].mask(precision_df['precision_bp'].isna(), inplace=True)
        precision_df.drop(columns=['cumsum_length_pct', 'total_length_pct'], inplace=True)
    if self.options.genome_to_unique_common:
        precision_df = precision_df[~precision_df['genome_id'].isin(self.options.genome_to_unique_common)]"""

    precision_avg_bp = precision_df["precision_bp"].mean()
    precision_avg_bp_sem = precision_df["precision_bp"].sem()
    precision_avg_bp_var = precision_df["precision_bp"].var()
    precision_avg_seq = precision_df["precision_seq"].mean()
    precision_avg_seq_sem = precision_df["precision_seq"].sem()
    precision_weighted_bp = precision_df["tp_length"].sum() / precision_df["total_length"].sum()
    precision_weighted_seq = precision_df["tp_seq_counts"].sum() / precision_df["total_seq_counts"].sum()

    genome_sizes_df = (
        gs_df.groupby("genome_id", sort=False)
        .agg({"seq_length": "sum", "SEQUENCEID": "count"})
        .rename(columns={"seq_length": "length_gs", "SEQUENCEID": "seq_counts_gs"})
    )
    precision_df = (
        precision_df.reset_index().join(genome_sizes_df, on="genome_id", how="left", sort=False).set_index("BINID")
    )
    precision_df["recall_bp"] = precision_df["tp_length"] / precision_df["length_gs"]
    precision_df["recall_seq"] = precision_df["tp_seq_counts"] / precision_df["seq_counts_gs"]
    precision_df["rank"] = "NA"

    recall_df = confusion_df.loc[confusion_df.groupby("genome_id", sort=False)["genome_length"].idxmax()]
    recall_df = (
        recall_df.reset_index().join(genome_sizes_df, on="genome_id", how="right", sort=False).set_index("BINID")
    )
    recall_df.fillna({"genome_length": 0, "genome_seq_counts": 0}, inplace=True)
    recall_df["recall_bp"] = recall_df["genome_length"] / recall_df["length_gs"]
    recall_df["recall_seq"] = recall_df["genome_seq_counts"] / recall_df["seq_counts_gs"]

    recall_df = recall_df.join(precision_df[["total_length", "seq_length_mean"]], how="left", sort=False)

    # if self.options.genome_to_unique_common:
    #    recall_df = recall_df[~recall_df["genome_id"].isin(self.options.genome_to_unique_common)]

    recall_avg_bp = recall_df["recall_bp"].mean()
    recall_avg_bp_var = recall_df["recall_bp"].var()
    recall_avg_bp_sem = recall_df["recall_bp"].sem()
    recall_avg_seq = recall_df["recall_seq"].mean()
    recall_avg_seq_sem = recall_df["recall_seq"].sem()
    recall_weighted_bp = recall_df["genome_length"].sum() / recall_df["length_gs"].sum()
    recall_weighted_seq = recall_df["genome_seq_counts"].sum() / recall_df["seq_counts_gs"].sum()

    # Compute recall as in CAMI 1
    """unmapped_genomes = set(gs_df["genome_id"].unique()) - set(precision_df["genome_id"].unique())
    #if self.options.genome_to_unique_common:
    #    unmapped_genomes -= set(self.options.genome_to_unique_common)
    num_unmapped_genomes = len(unmapped_genomes)
    prec_copy = precision_df.reset_index()
    if num_unmapped_genomes:
        prec_copy = prec_copy.reindex(
            prec_copy.index.tolist() + list(range(len(prec_copy), len(prec_copy) + num_unmapped_genomes))
        ).fillna(0.0)
    self.metrics.recall_avg_bp_cami1 = prec_copy["recall_bp"].mean()
    self.metrics.recall_avg_seq_cami1 = prec_copy["recall_seq"].mean()
    self.metrics.recall_avg_bp_sem_cami1 = prec_copy["recall_bp"].sem()
    self.metrics.recall_avg_seq_sem_cami1 = prec_copy["recall_seq"].sem()
    self.metrics.recall_avg_bp_var_cami1 = prec_copy["recall_bp"].var()
    self.recall_df_cami1 = prec_copy"""
    # End Compute recall as in CAMI 1

    accuracy_bp = precision_df["tp_length"].sum() / recall_df["length_gs"].sum()
    accuracy_seq = precision_df["tp_seq_counts"].sum() / recall_df["seq_counts_gs"].sum()

    precision_df = precision_df.sort_values(by=["recall_bp"], axis=0, ascending=False)
    recall_df = recall_df
    metrics = {
        "precision_avg_bp": precision_avg_bp,
        "precision_avg_bp_sem": precision_avg_bp_sem,
        "precision_avg_bp_var": precision_avg_bp_var,
        "precision_avg_seq": precision_avg_seq,
        "precision_avg_seq_sem": precision_avg_seq_sem,
        "precision_weighted_bp": precision_weighted_bp,
        "precision_weighted_seq": precision_weighted_seq,
        "recall_avg_bp": recall_avg_bp,
        "recall_avg_bp_sem": recall_avg_bp_sem,
        "recall_avg_bp_var": recall_avg_bp_var,
        "recall_avg_seq": recall_avg_seq,
        "recall_avg_seq_sem": recall_avg_seq_sem,
        "recall_weighted_bp": recall_weighted_bp,
        "recall_weighted_seq": recall_weighted_seq,
        "accuracy_bp": accuracy_bp,
        "accuracy_seq": accuracy_seq,
        "f1_avg_bp": (2*precision_avg_bp*recall_avg_bp)/(precision_avg_bp+recall_avg_bp) if (precision_avg_bp+recall_avg_bp) > 0 else 0
    }
    bins_eval = calc_num_recovered_genomes(precision_df, [0.9, 0.5], [0.05, 0.1])
    return metrics, bins_eval


def calc_num_recovered_genomes(bins, min_completeness, max_contamination):
    counts_list = []
    for x in itertools.product(min_completeness, max_contamination):
        count = bins[(bins["recall_bp"] > x[0]) & (bins["precision_bp"] > (1 - x[1]))].shape[0]
        counts_list.append(("> " + str(int(x[0] * 100)) + "% completeness", "< " + str(int(x[1] * 100)) + "%", count))

    pd_counts = pd.DataFrame(counts_list, columns=["Completeness", "Contamination", "count"])
    pd_counts = pd.pivot_table(
        pd_counts,
        values="count",
        index=["Contamination"],
        columns=["Completeness"],
    ).reset_index()
    return pd_counts



from os import uname
import tensorflow as tf

from tensorflow.keras.optimizers import Adam, SGD
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Dense, Activation, Softmax, Embedding, LayerNormalization
#from tensorflow.keras import Sequential
from tensorflow.keras.layers import BatchNormalization, Lambda, LeakyReLU, Flatten
from tensorflow.keras.layers import Dropout, Add, Concatenate
from tensorflow.keras.losses import SparseCategoricalCrossentropy
import numpy as np
from tensorflow.keras.regularizers import l2    
from tqdm import tqdm

#import tensorflow_probability as tfp

class VAEEncoder(Model):
    def __init__(self, abundance_dim, kmers_dim, hiddendim, zdim=64, dropout=0, layers=2):
        super(VAEEncoder, self).__init__()
        self.abundance_dim = abundance_dim
        self.kmers_dim = kmers_dim
        self.zdim = zdim
        in_ = Input(shape=(abundance_dim+kmers_dim,))
        x = in_
        for i in range(layers):
            x = Dense(hiddendim, activation='linear', name=f"encoder_{i}")(x)
            x = LeakyReLU(0.01)(x)
            if dropout > 0:
                x = Dropout(dropout)(x)
            x = BatchNormalization()(x)
        mu = Dense(zdim, name="mu")(x)
        logvar = Dense(zdim, kernel_initializer='zeros', name="logvar")(x)
        self.model = Model(in_, [mu, logvar])
 
    def call(self, x, training=False):
        mu, sigma = self.model(x, training=training)
        return mu, sigma

    
class VAEDecoder(Model):
    def __init__(self, abundance_dim, kmers_dim, hiddendim, zdim=64, dropout=0, layers=2):
        super(VAEDecoder, self).__init__()
        self.abundance_dim = abundance_dim
        self.kmers_dim = kmers_dim
        self.zdim = zdim
        in_ = Input(shape=(zdim,))
        x = in_
        for i in range(layers):
            x = Dense(hiddendim, activation='linear', name=f"decoder_{i}")(x)
            x = LeakyReLU(0.01)(x)
            if dropout > 0:
                x = Dropout(dropout)(x)
            x = BatchNormalization()(x)
            
        x1 = Dense(abundance_dim, name="decoder_ab")(x)
        x2 = Dense(kmers_dim, name="decoder_kmer")(x)
        if self.abundance_dim > 1:
            x1 = Softmax()(x1)
        x = Concatenate()((x1,x2))
        
        self.model = Model(in_, x)
 
    def call(self, z, training=False):
        x_hat = self.model(z, training=training)
        return x_hat

class Encoder(Model):
    def __init__(self, features_dim, hiddendim, zdim=64, dropout=0, layers=2):
        super(VAEEncoder, self).__init__()
        self.zdim = zdim
        in_ = Input(shape=(features_dim,))
        x = in_
        for i in range(layers):
            x = Dense(hiddendim, activation='linear', name=f"encoder_{i}")(x)
            x = LeakyReLU(0.01)(x)
            if dropout > 0:
                x = Dropout(dropout)(x)
            x = BatchNormalization()(x)
        mu = Dense(zdim, name="mu")(x)
        logvar = Dense(zdim, kernel_initializer='zeros', name="logvar")(x)
        self.model = Model(in_, [mu, logvar])
 
    def call(self, x, training=False):
        mu, sigma = self.model(x, training=training)
        return mu, sigma

    
class Decoder(Model):
    def __init__(self, features_dim, hiddendim, zdim=64, dropout=0, layers=2):
        super(VAEDecoder, self).__init__()
        self.features_dim = features_dim
        self.zdim = zdim
        in_ = Input(shape=(zdim,))
        x = in_
        for i in range(layers):
            x = Dense(hiddendim, activation='linear', name=f"decoder_{i}")(x)
            x = LeakyReLU(0.01)(x)
            if dropout > 0:
                x = Dropout(dropout)(x)
            x = BatchNormalization()(x)
            
        x = Dense(features_dim, name="decoder_ab")(x)
        #x2 = Dense(kmers_dim, name="decoder_kmer")(x)
        #if self.abundance_dim > 1:
        #    x1 = Softmax()(x1)
        #x = Concatenate()((x1,x2))
        
        self.model = Model(in_, x)
 
    def call(self, z, training=False):
        x_hat = self.model(z, training=training)
        return x_hat


class LabelClassifier(Model):
    def __init__(self, n_classes, hiddendim=128, zdim=64, dropout=0, layers=1):
        super(LabelClassifier, self).__init__()
        in_ = Input(shape=(zdim,))
        x = in_
        for nl in range(layers):
            if nl == layers-1:
                x = Dense(n_classes, activation='softmax')(x)
            else:
                x = Dense(hiddendim, activation='linear')(x)
                #x = LeakyReLU(0.01)(x)
                if dropout > 0:
                    x = Dropout(dropout)(x)

        self.model = Model(in_, x)
        self.loss_fn = SparseCategoricalCrossentropy()
 
    def call(self, z, mask, training=False):
        predictions = self.model(tf.gather(params=z, indices=mask), training=training)
        return predictions

    def loss(self, gold_labels, predicted_labels):
        return self.loss_fn(y_true=gold_labels,y_pred=predicted_labels)

class TrainHelperVAE:
    def __init__(self, encoder, decoder, learning_rate=1e-3,  kld_weight=1/200.,
                train_weights=False, classification=False, n_classes=0,
                gold_labels=None, mask_labels=0.0, ae_alpha=1):
        self.encoder = encoder
        self.decoder = decoder
        self.train_weights = train_weights
        self.abundance_dim = self.encoder.abundance_dim
        self.kmers_dim = self.encoder.kmers_dim
        self.ae_alpha = ae_alpha
        if train_weights:
            #self.kld_weight = tf.Variable(kld_weight)
            self.kld_weight = kld_weight
            #self.abundance_weight = tf.Variable(0.5)
            self.kmer_weight = tf.Variable(0.1)
        else:
            self.kld_weight = kld_weight
            if self.abundance_dim > 1:
                #self.abundance_weight = 0.85 / tf.math.log(float(self.abundance_dim))
                self.kmer_weight = 0.15
            else:
                #self.abundance_weight = 0.5
                self.kmer_weight = 0.5
        self.opt = tf.keras.optimizers.Adam(learning_rate=learning_rate)
        self.logvar = None
        self.mu = None
        self.z = None
        self.classify = classification
        if self.classify:
            self.classifier = LabelClassifier(n_classes, zdim=encoder.zdim)
            self.gold_labels = gold_labels
            self.mask_labels = mask_labels
            
    def train_step(self, x, writer=None, epoch=0, vae=True, gold_labels=None):
        losses = self._train_step(x, vae=vae, writer=writer, epoch=epoch, gold_labels=gold_labels)
        if writer is not None:
            with writer.as_default():
                tf.summary.scalar('mean logvar', self.logvar, step=epoch)
                tf.summary.scalar('mean mu', self.mu, step=epoch)
        losses = [loss.numpy() for loss in losses]
        return losses
    
    @tf.function
    def loss(self, x, mu, logvar, vae, training=True, writer=None, epoch=0, gold_labels=None):
        if vae:
            epsilon = tf.random.normal(tf.shape(mu))
            z = mu + epsilon * tf.math.exp(0.5 * logvar)
            kld  = 0.5*tf.math.reduce_mean(tf.math.reduce_mean(1.0 + logvar - tf.math.pow(mu, 2) - tf.math.exp(logvar), axis=1))
            kld  = kld * self.kld_weight
        else:
            z = mu
            kld = tf.convert_to_tensor(0.0)
        x_hat = self.decoder(z, training=training)
        if self.classify:
            if self.mask_labels > 0:
                use_labels = np.random.choice(z.shape[0], int(z.shape[0]*(1-self.mask_labels)))
            else:
                use_labels = np.arange(z.shape[0])
            predictions = self.classifier(z, mask=use_labels)
            prediction_loss = self.classifier.loss(gold_labels[use_labels], predictions)
        else:
            prediction_loss = tf.convert_to_tensor(0.0)
        if writer is not None:
            with writer.as_default():
                tf.summary.scalar('min ab', tf.reduce_min(x_hat[:, :self.abundance_dim]), step=epoch)
        if self.abundance_dim > 1:
            mse1 = -tf.reduce_mean(tf.reduce_sum((tf.math.log(x_hat[:, :self.abundance_dim] + 1e-9) * x[:, :self.abundance_dim]), axis=1))
        else:
            mse1 = tf.reduce_mean( (x[:, :self.abundance_dim] - x_hat[:, :self.abundance_dim])**2)
        if self.train_weights:
            kmer_weight = tf.math.sigmoid(self.kmer_weight)
        else:
            kmer_weight = self.kmer_weight
        mse1 *= (1-kmer_weight)
        mse2 = kmer_weight*tf.reduce_mean( tf.reduce_mean((x[:, self.abundance_dim:] - x_hat[:, self.abundance_dim:])**2, axis=1))

        return mse1, mse2, kld, prediction_loss

    @tf.function
    def _train_step(self, x, vae=True, writer=None, epoch=0, gold_labels=None):
        with tf.GradientTape() as tape:
            mu, logvar = self.encoder(x, training=True)
            logvar = tf.clip_by_value(logvar, -2, 2)
            self.logvar = tf.cast(tf.math.reduce_mean(logvar), float)
            self.mu = tf.cast(tf.math.reduce_mean(mu), float)
            mse1, mse2, kld, predl = self.loss(x, mu, logvar, vae, training=True,
                                               writer=writer, epoch=epoch,
                                               gold_labels=gold_labels)
            loss = mse1 + mse2 - kld + predl
            loss *= self.ae_alpha
        tw = self.encoder.trainable_weights + self.decoder.trainable_weights 
        if self.train_weights:
            tw += [self.kmer_weight] # self.kld_weight
        if self.classify:
            tw += self.classifier.trainable_weights
        grads = tape.gradient(loss, tw)
        grad_norm = tf.linalg.global_norm(grads)
        clip_grads, _ = tf.clip_by_global_norm(grads, 5,  use_norm=grad_norm)
        new_grad_norm = tf.linalg.global_norm(clip_grads)
        self.opt.apply_gradients(zip(clip_grads, tw))

        if writer is not None:
            with writer.as_default():
                tf.summary.scalar('grad norm', grad_norm, step=epoch)
                tf.summary.scalar('clipped grad norm', new_grad_norm, step=epoch)

        return loss, mse2, mse1, kld, predl


class NoiseModel(Model):
    def __init__(self, abundance_dim, emb_dim, hiddendim=128, dropout=0, layers=1):
        super(NoiseModel, self).__init__()
        self.abundance_dim = abundance_dim
        self.kmers_dim = emb_dim
        in_ = Input(shape=((abundance_dim+emb_dim)*2,))
        x = in_
        for i in range(layers):
            x = Dense(hiddendim, activation='linear', name=f"noise_{i}")(x)
            x = LeakyReLU(0.01)(x)
            if dropout > 0:
                x = Dropout(dropout)(x)
            x = BatchNormalization()(x)
        noise = Dense(1, name="mu")(x)
        self.model = Model(in_, noise)
 
    def call(self, x, training=False):
        noise = self.model(x, training=training)
        return noise

class TH:
    def __init__(
        self,
        input_features,
        gnn_model,
        lr=0.01,
        all_different_idx=None,
        all_same_idx=None,
        ae_encoder=None,
        ae_decoder=None,
        decoder_input="gnn",
        classifier=None,
        latentdim=32,
        gnn_weight=1.0,
        ae_weight=1.0,
        kld_weight = 1/200,
        scg_weight=100.0,
        kmer_weight=0.15,
        abundance_weight=0.85,
        kmer_dim=103,
        kmer_alpha=0.5,
        num_negatives=50,
        kmers_dim=103,
        abundance_dim=4,
        labels=None,
        use_gnn=True,
        use_noise=False,
        loglevel="warning",
        pretrainvae=0
    ):
        self.opt = Adam(learning_rate=lr, epsilon=1e-8)
        # self.opt = SGD(learning_rate=lr)
        self.gnn_model = gnn_model
        self.features = input_features
        # self.dense_adj = tf.sparse.to_dense(self.model.adj)
        if gnn_model is not None:
            self.adj_shape = self.gnn_model.adj.dense_shape
        self.kmer_dim = kmer_dim
        self.ab_dim = input_features.shape[1] - kmer_dim
        self.kmer_alpha = kmer_alpha
        self.num_negatives = num_negatives
        self.nlatent = latentdim
        self.scg_pairs = all_different_idx
        self.all_same_idx = all_same_idx
        self.classifier = classifier
        self.encoder = ae_encoder
        self.decoder = ae_decoder
        self.decoder_input = decoder_input
        self.use_ae = ae_encoder is not None
        self.mse_loss = tf.keras.losses.MeanSquaredError()
        self.gnn_weight = gnn_weight
        self.ae_weight = ae_weight
        self.kld_weight = kld_weight
        self.scg_weight = scg_weight
        self.kmer_weight = kmer_weight
        self.noise_reg = 1
        self.abundance_weight = abundance_weight
        self.no_gnn = gnn_model is None
        self.train_ae = False
        self.abundance_dim = abundance_dim
        self.kmers_dim = kmers_dim
        self.use_gnn = use_gnn
        self.use_noise = use_noise
        self.loglevel = loglevel
        self.pretrainvae = pretrainvae
        self.epoch = 0
        if self.use_noise:
            #noise_units = 1
            #self.positive_noises = tf.Variable(tf.random_normal_initializer()(shape=(self.gnn_model.adj.values.shape + noise_units), dtype=tf.float32), trainable=True)
            #self.scg_noises = tf.Variable(tf.random_normal_initializer()(shape=(self.scg_pairs.shape[0], noise_units), dtype=tf.float32), trainable=True)
            #self.noise_weights = tf.Variable(tf.random_normal_initializer()(shape=(noise_units,), dtype=tf.float32), trainable=True)
            # alt: Dense layer with concat of embs (or feats) as input and single value output
            self.noise_model = NoiseModel(abundance_dim=abundance_dim, emb_dim=0)
            self.noise_dict = {}

    def sample_negatives_old(self, edge_idx):
        # get numbers between 0 and n*(n-1)
        neg_idx = tf.random.uniform(
            shape=[len(edge_idx)*self.num_negatives,],
            minval=0,
            maxval=self.adj_shape[0] * self.adj_shape[1] - 1,
            dtype=tf.int64,
        )
        neg_idx_row = tf.math.minimum(
            tf.cast(self.adj_shape[0] - 1, tf.float32),
            tf.cast(neg_idx, tf.float32) / tf.cast(self.adj_shape[1], tf.float32),
        )
        neg_idx_row = tf.cast(neg_idx_row, tf.int64)
        neg_idx_col = tf.cast(tf.math.minimum(self.adj_shape[1] - 1,
                              (neg_idx % self.adj_shape[1])), tf.int64)
        return neg_idx_row, neg_idx_col

    @tf.function
    def sample_negatives(self, edge_idx, node_idx):
        # for 5 negatives and batchsize 256, this should give 1% or less false negatives
        neg_idx = tf.random.uniform(shape=(2, len(edge_idx)*self.num_negatives),
                                    maxval=node_idx.shape[0], dtype=tf.int32)
        # get neg idx relative to node_idx (of this batch)
        neg_idx_row = tf.gather(params=node_idx, indices=neg_idx[0,:])
        neg_idx_col = tf.gather(params=node_idx, indices=neg_idx[1,:])
        return neg_idx_row, neg_idx_col
        #return neg_idx[0,:], neg_idx[1,:]

    @tf.function
    def nodedist(self, u, v):
        #breakpoint()
        #return tf.reduce_sum((tf.expand_dims(x, 0) - tf.expand_dims(y, 1)**2), axis=-1)
        u = tf.nn.l2_normalize(u, axis=1)
        v = tf.nn.l2_normalize(v, axis=1)
        pairwise = tf.reduce_sum(tf.math.multiply(u, v), axis=1)
        #pairwise = -tf.norm(tf.math.subtract(u, v) + + 1.0e-12, ord='euclidean', axis=1,)
        return pairwise
    
    @tf.function
    def edge_loss(self, pos_dists, neg_dists):
        #breakpoint()
        y_true = tf.concat((tf.ones_like(pos_dists), tf.zeros_like(neg_dists)), axis=0)
        y_pred = tf.concat((pos_dists, neg_dists), axis=0)
        gnn_loss = tf.keras.metrics.binary_crossentropy(y_true, y_pred, from_logits=True)
        return gnn_loss
    
    def train_vae(self, nodes_idx, vae=True):
        if not vae: # not variational, no kld loss
            # make logvar non trainable
            layer_names = [layer.name for layer in self.encoder.layers[0].layers]
            logvar_idx = layer_names.index("logvar")
            self.encoder.layers[0].layers[logvar_idx].trainable = False
            #ae_embs = tf.concat((self.features[:,:self.abundance_dim], self.encoder(self.features)[0]), axis=1)

        ae_mu, ae_logvar = self.encoder(tf.gather(self.features, nodes_idx), training=(self.epoch < self.pretrainvae or self.pretrainvae == 0))
        #ae_mu, ae_logvar = self.encoder(tf.gather(self.features, nodes_idx), training=True)
        if self.decoder_input == "gnn" and self.use_gnn and (self.epoch > self.pretrainvae or self.pretrainvae == 0):
            ae_mu = tf.scatter_nd(indices=nodes_idx[:,None],
                                     updates=ae_mu,
                                     shape=(self.features.shape[0], ae_mu.shape[1]))
            ae_mu = self.gnn_model(ae_mu, nodes_idx, training=True)
        ae_recon = self.decoder(ae_mu, training=(self.epoch < self.pretrainvae or self.pretrainvae == 0))
        #ae_recon = self.decoder(ae_mu, training=True)
        ae_logvar = tf.clip_by_value(ae_logvar, -2, 2)
        try:
            losses = self.ae_loss(tf.gather(self.features, nodes_idx), ae_recon, ae_mu, ae_logvar, vae=vae)
        except:
            breakpoint()
        ae_loss = tf.reduce_sum(losses) * self.ae_weight
        ae_losses = {"kmer_loss": losses[0],
                        "ab_loss": losses[1],
                        "kld": losses[2],
                        "vae_loss": ae_loss,
                        "mean_logvar": tf.reduce_mean(ae_logvar)}
        if ae_losses["kld"] < -1:
            breakpoint()
        #logvar_idx = layer_names.index("logvar")
        #self.encoder.layers[0].layers[logvar_idx].trainable = False
        #ae_embs = tf.concat((self.features[:,:self.abundance_dim], self.encoder(self.features)[0]), axis=1)
        #ae_embs = self.encoder(self.features)[0]
        return ae_mu, ae_losses

    def train_edges(self, node_hat, nodes_idx, edges_idx, train_pairs):
        """_summary_

        :param node_hat: node embeddings
        :type node_hat: _type_
        :param nodes_idx: node idx to be considered in this batch
        :type nodes_idx: _type_
        :param edges_idx: edge idx to be considered in this batch
        :type edges_idx: _type_
        :param train_pairs: graph edges, pairs of node idxs of this batch
        :type train_pairs: 
        :return: _description_
        :rtype: _type_
        """
        gnn_losses = {"gnn_loss": tf.constant(0, dtype=tf.float32),
                      "pos": tf.constant(0, dtype=tf.float32),
                      "neg": tf.constant(0, dtype=tf.float32)}
        if self.gnn_weight > 0:
            #breakpoint()
            src_embs = tf.gather(indices=train_pairs[0], params=node_hat)
            dst_embs = tf.gather(indices=train_pairs[1], params=node_hat)
            positive_pairwise_dist = self.nodedist(src_embs, dst_embs)
            # create random negatives for gnn_loss
            batch_neg_idx_src, batch_neg_idx_dst = self.sample_negatives(edge_idx=edges_idx,
                                                                node_idx=nodes_idx)
            if self.loglevel == "debug":
                pset = set(zip(train_pairs[0].numpy(), train_pairs[1].numpy()))
                nset = set(zip(batch_neg_idx_src.numpy(), batch_neg_idx_dst.numpy()))
                print("false random negatives", round(len(pset & nset)/len(edges_idx), 4))
            
            try:
                #negative_pairs = tf.gather_nd(pairwise_similarity, neg_idx)
                neg_row_embs = tf.gather(indices=batch_neg_idx_src, params=node_hat)
                neg_col_embs = tf.gather(indices=batch_neg_idx_dst, params=node_hat)
                negative_pairwise_dist = self.nodedist(neg_row_embs, neg_col_embs)
            except:
                breakpoint()
            if self.use_noise:
                # reduce noises to one dim
                src_ab = tf.gather(indices=train_pairs[0], params=self.features )[:, :self.ab_dim]
                dst_ab = tf.gather(indices=train_pairs[1], params=self.features )[:, :self.ab_dim]
                noise_input = tf.concat((src_ab, dst_ab), axis=1)
                pos_noises = self.noise_model(noise_input)[:,0]
                positive_pairwise_dist = positive_pairwise_dist + pos_noises

                src_ab = tf.gather(indices=batch_neg_idx_src, params=self.features )[:, :self.ab_dim]
                dst_ab = tf.gather(indices=batch_neg_idx_dst, params=self.features )[:, :self.ab_dim]
                noise_input = tf.concat((src_ab, dst_ab), axis=1)
                neg_noises = self.noise_model(noise_input)[:,0]
                negative_pairwise_dist = negative_pairwise_dist + neg_noises
                self.noise_dict.update({(train_pairs[0][i].numpy(), train_pairs[1][i].numpy()): pos_noises[i].numpy() for i in range(len(edges_idx))})
                self.noise_dict.update({(batch_neg_idx_src[i].numpy(), batch_neg_idx_dst[i].numpy()): neg_noises[i].numpy() for i in range(len(batch_neg_idx_src))})



            pos_dist = tf.reduce_mean(positive_pairwise_dist)
            if self.num_negatives > 0:
                neg_dist = tf.reduce_mean(negative_pairwise_dist)
            else:
                neg_dist = tf.constant(0)
            #gnn_loss = (pos_loss + neg_loss) * self.gnn_weight
            #neg_loss = 0
            gnn_losses["pos"] = pos_dist
            gnn_losses["neg"] = neg_dist
            gnn_loss = self.edge_loss(positive_pairwise_dist, negative_pairwise_dist)
            gnn_loss = gnn_loss * self.gnn_weight
            gnn_losses["gnn_loss"] = gnn_loss
        #loss = gnn_loss
        return gnn_losses

    def train_scg(self, node_hat, scgs_idx):
        scg_loss = tf.constant(0, dtype=tf.float32)
        if self.scg_pairs is not None and self.scg_weight > 0:
            scg_row_embs = tf.gather(node_hat, self.scg_pairs[scgs_idx, 0])
            scg_col_embs = tf.gather(node_hat, self.scg_pairs[scgs_idx, 1])
            scg_pairwise = self.nodedist(scg_row_embs, scg_col_embs)
            #scg_pairwise = tf.nn.sigmoid(scg_pairwise)
            #if self.use_noise:
            #    scg_noises = tf.matmul(tf.gather(indices=scgs_idx, params=self.scg_noises), noises_weights)
            #    scg_pairwise = scg_pairwise + scg_noises[:,0]
                #scg_pairwise = tf.clip_by_value(scg_pairwise, 0, 1)
            #scg_loss = tf.keras.losses.binary_crossentropy(
            #        tf.zeros_like(scg_pairwise), scg_pairwise, from_logits=True
            #)
            scg_loss = self.edge_loss(tf.ones_like([0.0]), scg_pairwise)
            scg_loss *= self.scg_weight
        return scg_loss

    @tf.function
    def train_scg_only(self):
        scgs_idx = range(0, len(self.scg_pairs))
        layer_names = [layer.name for layer in self.encoder.layers[0].layers]
        logvar_idx = layer_names.index("logvar")
        self.encoder.layers[0].layers[logvar_idx].trainable = False
        with tf.GradientTape() as tape:

            node_hat, _ = self.encoder(self.features, training=True)
            scg_loss = self.train_scg(node_hat, scgs_idx)
            # aggregate model weights
            if self.use_gnn and (self.epoch > self.pretrainvae or self.pretrainvae ==  0) :
                tw = self.gnn_model.trainable_weights
            else:
                tw = []
            if self.use_ae:
                if self.pretrainvae == 0 or self.epoch < self.pretrainvae:
                    tw += self.encoder.trainable_weights
                #self.encoder.layers[0].layers[logvar_idx].trainable = True
            grads = tape.gradient(scg_loss, tw)

            self.opt.apply_gradients(zip(grads, tw))
        self.encoder.layers[0].layers[logvar_idx].trainable = True
        return scg_loss


    @tf.function
    def train_unsupervised(self, nodes_idx=None, edges_idx=None,
                            scgs_idx=None,training=True, vae=True,
                            mask_labels=None, gold_labels=None, last_batch=False, scg=False):
        #### get node indices to be used for this batch
        if edges_idx is None:
            edges_idx = range(0,self.gnn_model.adj.indices.shape[0])
        if nodes_idx is None:
            # get nodes_idx from edges_idx
            # this are the node pairs with their original indices, e.g. (1,2), (2,3), (3,1)
            train_src_original = tf.gather(indices=edges_idx, params=self.gnn_model.adj.indices[:,0])
            train_dst_original = tf.gather(indices=edges_idx, params=self.gnn_model.adj.indices[:,1])
            unique_nodes = tf.unique(tf.concat((train_src_original, train_dst_original), axis=0))
            nodes_idx = unique_nodes.y # e.g. (A,B,C)
            # get new indices for edges in relation to current node list(A,B), (B,C), (C,A)
            #train_src_new = unique_nodes.idx[:train_src_original.shape[0]] 
            #train_dst_new = unique_nodes.idx[train_src_original.shape[0]:]
            #train_idx_new = (train_src_new, train_dst_new)
            if self.loglevel == "debug":
                print(f"using {nodes_idx.shape} nodes for this batch")
        else:
            train_idx_new = (self.gnn_model.adj.indices[:,0], self.gnn_model.adj.indices[:,1])
            train_src_original = self.gnn_model.adj.indices[:,0]
            train_dst_original = self.gnn_model.adj.indices[:,1]
        if scgs_idx is None:
            scgs_idx = range(0, len(self.scg_pairs))
        #####
        
        with tf.GradientTape() as tape:
            #####   run encoder first on nodes of this batch
            if self.use_ae:
                # only nodes in nodes_idx are processed, the output may have a different dimension
                ae_embs, ae_losses = self.train_vae(nodes_idx)
                # ae_embs is only nodes in node_idx
                #reverse gather, expand so that ae_embs has the same dim as self.features
                ae_embs = tf.scatter_nd(indices=nodes_idx[:,None],
                                     updates=ae_embs,
                                     shape=(self.features.shape[0], ae_embs.shape[1]))
            else:
                ae_embs = self.features # ae_embs is all nodes
                ae_losses = {}
            ######
            ###### run gnn model
            if self.use_gnn and self.decoder_input == "vae" and (self.epoch > self.pretrainvae or self.pretrainvae ==  0):
                gnn_embs = self.gnn_model(ae_embs, nodes_idx, training=True)
                node_hat = tf.scatter_nd(indices=nodes_idx[:,None],
                                     updates=gnn_embs,
                                     shape=(self.features.shape[0], gnn_embs.shape[1]))
            else:
                node_hat = ae_embs
            gnn_losses = self.train_edges(node_hat, nodes_idx, edges_idx,
                                          (train_src_original, train_dst_original))
            scg_loss = tf.convert_to_tensor(0.0)
            if scg:
                scg_loss = self.train_scg(node_hat, scgs_idx)
            gnn_losses["scg_loss"] = scg_loss

            # classification loss
            if self.classifier is not None:
                if mask_labels is not None and mask_labels > 0:
                    use_labels = np.random.choice(node_hat.shape[0], int(node_hat.shape[0]*(1-self.mask_labels)))
                else:
                    use_labels = np.arange(node_hat.shape[0])
                predictions = self.classifier(node_hat, mask=use_labels)
                pred_loss = self.classifier.loss(gold_labels[use_labels], predictions)
            else:
                pred_loss = tf.convert_to_tensor(0.0)
            gnn_losses["pred_loss"] = pred_loss
            
            # combine losses and update model
            loss = gnn_losses["gnn_loss"] + scg_loss
            if training:
                # aggregate model weights
                if self.use_gnn and (self.epoch > self.pretrainvae or self.pretrainvae ==  0) :
                    tw = self.gnn_model.trainable_weights
                else:
                    tw = []
                if self.use_ae:
                    loss += ae_losses["vae_loss"]
                    if self.pretrainvae == 0 or self.epoch < self.pretrainvae:
                        tw += self.encoder.trainable_weights + self.decoder.trainable_weights # skip logvar 
                    #self.encoder.layers[0].layers[logvar_idx].trainable = True
                if self.use_noise:
                    #breakpoint()
                    #tw += [self.positive_noises, self.scg_noises]
                    tw += self.noise_model.trainable_weights
                    #loss += noise_loss
                if self.classifier is not None:
                    tw += self.classifier.trainable_weights
                    loss += pred_loss
                #################
                #grads = tape.gradient([ae_losses.get("vae_loss", tf.convert_to_tensor(0.0)), gnn_losses["gnn_loss"], scg_loss, pred_loss], tw)
                grads = tape.gradient([ae_losses.get("vae_loss", tf.convert_to_tensor(0.0)), gnn_losses["gnn_loss"], pred_loss], tw)
                grad_norm = tf.linalg.global_norm(grads)
                #clip_grads, _ = tf.clip_by_global_norm(grads, 2,  use_norm=grad_norm)
                #new_grad_norm = tf.linalg.global_norm(clip_grads)
                self.opt.apply_gradients(zip(grads, tw))
                ae_losses["grad_norm"] = grad_norm.numpy()
                #ae_losses["grad_norm_clip"] = new_grad_norm.numpy()
        return loss, gnn_losses, ae_losses
        
        
    #@tf.function
    def ae_loss(self, x, x_hat, mu, logvar, vae):
        if vae:
            epsilon = tf.random.normal(tf.shape(mu))
            z = mu + epsilon * tf.math.exp(0.5 * logvar)
            kld  = 0.5*tf.math.reduce_mean(tf.math.reduce_mean(1.0 + logvar - tf.math.pow(mu, 2) - tf.math.exp(logvar), axis=1))
            kld  = -kld * self.kld_weight
        else:
            z = mu
            kld = tf.convert_to_tensor(0.0)
        if self.abundance_dim > 1:
            mse1 = - tf.reduce_mean(tf.reduce_sum((tf.math.log(x_hat[:, :self.abundance_dim] + 1e-9) * x[:, :self.abundance_dim]), axis=1))
        else:
            mse1 = tf.reduce_mean( (x[:, :self.abundance_dim] - x_hat[:, :self.abundance_dim])**2)
        mse1 *= (1-self.kmer_weight)
        mse2 = self.kmer_weight*tf.reduce_mean( tf.reduce_mean((x[:, self.abundance_dim:] - x_hat[:, self.abundance_dim:])**2, axis=1))

        return mse2, mse1, kld

    @tf.function
    def train_unsupervised_decode(self, idx):
        with tf.GradientTape() as tape:
            # run gnn model
            z_sample, mu, logvar, x_orginal, x_hat = self.gnn_model(self.features, self.adj,
                                                           indices=idx, training=True)
 

            loss = 0
            gnn_loss = tf.constant(0, dtype=tf.float32)
            # SCG loss
            scg_loss = tf.constant(0, dtype=tf.float32)
            if self.scg_pairs is not None and self.scg_weight > 0:
                ns1 = tf.gather(z_sample, self.scg_pairs[:, 0])
                ns2 = tf.gather(z_sample, self.scg_pairs[:, 1])
                all_diff_pairs = tf.math.exp(-0.5 * tf.reduce_sum((ns1 - ns2) ** 2, axis=-1))
                scg_loss = tf.reduce_mean(all_diff_pairs) * self.scg_weight
                loss += scg_loss

            # decode
            kmer_loss, ab_loss, kld_loss = self.ae_loss(tf.gather(self.features, idx), x_hat, mu, logvar, vae=True)
            loss += kmer_loss + ab_loss - kld_loss

        tw = self.gnn_model.trainable_weights
        #tw += self.encoder.trainable_weights
        #tw += self.decoder.trainable_weights
        grads = tape.gradient(loss, tw)
        self.opt.apply_gradients(zip(grads, tw))
        return loss, gnn_loss, scg_loss, kmer_loss, ab_loss, kld_loss

    @staticmethod
    def sample_idx(idx, n):
        n = tf.cast(n, tf.int64)
        N = tf.cast(tf.shape(idx)[0], tf.int64)
        random_idx = tf.random.uniform(shape=(n,), minval=0, maxval=N, dtype=tf.int64)
        s_idx = tf.gather_nd(idx, random_idx)
        return s_idx




from setuptools import setup, find_packages
from setuptools.command.install import install
import os
import subprocess
from distutils.util import convert_path

main_ns = {}
ver_path = convert_path("src/graphmb/version.py")
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

setup(
    name="graphmb",
    version=main_ns["__version__"],
    packages=["graphmb"],
    package_dir={"": "src"},
    setup_requires=["setuptools~=58.0", "wheel", "sphinx-rtd-theme", "twine"],
    install_requires=[
        "wheel",
        "requests",
        "networkx==2.6.2",
        "torch==1.12.1",
        "scikit-learn==0.24.2",
        "tqdm==4.61.2",
        "pandas==1.3.5",
        "tensorflow==2.11.0",
        "tqdm==4.61.2",
        "numpy==1.23.5",
        "mlflow==2.1.1"

    ],
    entry_points={
        "console_scripts": ["graphmb=graphmb.main:main"],
    },
    include_package_data=True,
)




import networkx as nx
import numpy as np
#import tensorflow

colors = [
    "black",
    "red",
    "blue",
    "green",
    "orange",
    "purple",
    "brown",
    "pink",
    "yellow",
    "silver",
    "maroon",
    "fuchsia",
    "lime",
    "olive",
    "yellow",
    "navy",
    "teal",
    "steelblue",
    "darkred",
    "darkgreen",
    "darkblue",
    "darkorange",
    "lightpink",
    "lightgreen",
    "lightblue",
    "crimson",
    "darkviolet",
    "tomato",
    "tan",
    "tab:blue",
    "tab:orange",
    "tab:green",
    "tab:red",
    "tab:purple",
    "tab:brown",
    "tab:pink",
    "tab:gray",
    "tab:olive",
    "tab:cyan",
]

def plot_embs(node_ids, node_embeddings_2dim, labels_to_node, centroids, hq_centroids, node_sizes, outputname=None):
    """Plot embs of most labels with most support

    Args:
        node_ids ([type]): [description]
        node_embeddings_2dim ([type]): [description]
        labels_to_node ([type]): [description]
    """
    import matplotlib.pyplot as plt

    markers = ["o", "s", "p", "*"]
    if "NA" in labels_to_node:
        del labels_to_node["NA"]
    #breakpoint()
    labels_to_node = {label: labels_to_node[label] for label in labels_to_node if len(labels_to_node[label]) > 0}
    labels_to_plot = sorted(labels_to_node, key=lambda key: len(labels_to_node[key]), reverse=True)[
        : len(colors) * len(markers)
    ]#[:20]
    # print("ploting these labels", [l, colors[il], len(labels_to_node[l]) for il, l in enumerate(labels_to_plot)])
    x_to_plot = []
    y_to_plot = []
    colors_to_plot = []
    sizes_to_plot = []
    markers_to_plot = []
    #print(labels_to_plot)
    plt.figure()
    #print(" LABEL  COLOR  SIZE   DOTS")
    for i, l in enumerate(labels_to_plot):
        valid_nodes = 0
        if len(labels_to_node) == 0:
            continue
        for node in labels_to_node[l]:
            if node not in node_ids:
                # print("skipping", node)
                continue
            node_idx = node_ids.index(node)
            x_to_plot.append(node_embeddings_2dim[node_idx][0])
            y_to_plot.append(node_embeddings_2dim[node_idx][1])
            if node_sizes is not None:
                sizes_to_plot.append(node_sizes[node_idx])
            else:
                sizes_to_plot.append(50)
            valid_nodes += 1
            colors_to_plot.append(colors[i % len(colors)])
            markers_to_plot.append(markers[i // len(colors)])
        # breakpoint()
        #   print("plotting", l, colors[i % len(colors)], markers[i // len(colors)], len(labels_to_node[l]), valid_nodes)
        # plt.scatter(x_to_plot, y_to_plot, s=sizes_to_plot, c=colors[i], label=l)  # , alpha=0.5)
        sc = plt.scatter(
            x_to_plot,
            y_to_plot,
            s=sizes_to_plot,
            c=colors[i % len(colors)],
            label=l,
            marker=markers[i // len(colors)],
            alpha=0.4,
        )  # , alpha=0.5)
        x_to_plot = []
        y_to_plot = []
        sizes_to_plot = []
    plt.legend('')
    if centroids is not None:
        hq_centroids_mask = [x in hq_centroids for x in range(len(centroids))]
        lq_centroids_mask = [x not in hq_centroids for x in range(len(centroids))]
        # lq_centroids = set(range(len(centroids))) - set(hq_centroids)

        # plt.scatter(
        #    centroids[lq_centroids_mask, 0], centroids[lq_centroids_mask, 1], c="black", label="centroids (LQ)", marker="x"
        # )
        plt.scatter(
            centroids[hq_centroids_mask, 0],
            centroids[hq_centroids_mask, 1],
            c="black",
            label="centroids (HQ)",
            marker="P",
        )

    # for n in node_embeddings:
    # plt.scatter(x_to_plot, y_to_plot, c=colors_to_plot) #, alpha=0.5)
    #plt.legend()
    if outputname is not None:
        print("saving embs plot to {}".format(outputname))
        plt.savefig(outputname, bbox_inches="tight", dpi=400)
    else:
        plt.show()

def plot_edges_sim(X, adj, scgs, outname="", max_edge_value=150, min_edge_value=2):
    """
    X: feature matrix
    adj: adjacency matrix in sparse format
    """
    # for each pair in adj, calculate sim
    x_values = []
    y_values = []
    x_same_scgs = []
    y_same_scgs = []
    plotted_edges = set()
    for x, (i, j) in enumerate(zip(adj.row, adj.col)):
        if i != j and (i,j) not in plotted_edges and (j,i) not in plotted_edges and adj.data[x] > min_edge_value:
        #y_values.append(np.dot(X[i], X[j]))
            #y_values.append(scipy.spatial.distance.cosine(X[i], X[j]))
            plotted_edges.add((i,j))
            if len(scgs[i] & scgs[j]) > 0:
                y_same_scgs.append(np.dot(X[i], X[j])/(np.linalg.norm(X[i])*np.linalg.norm(X[j])))
                x_same_scgs.append(adj.data[x])
                #TODO plot edge weight by overlap
            else:
                y_values.append(np.dot(X[i], X[j])/(np.linalg.norm(X[i])*np.linalg.norm(X[j])))
                x_values.append(adj.data[x])
    if max_edge_value is not None:
        x_values = [min(x, max_edge_value) for x in x_values]
        x_same_scgs = [min(x, max_edge_value) for x in x_same_scgs]
    #x_values = adj.values
    #y_values = []
    #for (i, j) in adj.indices:
    #    y_values.append(np.dot(X[i], X[j]
    #))
    assert len(x_values) == len(y_values)
    import matplotlib.pyplot as plt
    plt.set_loglevel("error")
    
    plt.figure(0)
    plt.scatter(
            x_values,
            y_values, label=outname, marker=".", alpha=0.5, s=1)
    plt.scatter(
            x_same_scgs,
            y_same_scgs, label=outname+"SCG", marker="o", alpha=0.5, s=3)

    plt.xlabel("edge weight capped at {}".format(max_edge_value))

    plt.ylabel("cosine sim")
    plt.legend(loc='upper right')
    plt.savefig(outname + "edges_embs.png", dpi=500)
    #plt.show()
    plt.close()

    # dist histogram
    plt.figure(1)
    counts, edges, bars =   plt.hist(y_values, bins=50)
    plt.bar_label(bars)
    plt.savefig(outname + "embs_dists_histogram.png", dpi=500)
    plt.close()




def run_tsne(embs, dataset, cluster_to_contig, hq_bins, centroids=None):
    from sklearn.manifold import TSNE

    SEED = 0
    print("running tSNE")
    # filter only good clusters
    tsne = TSNE(n_components=2, random_state=SEED)
    if len(dataset.labels) == 1:
        label_to_node = {c: cluster_to_contig[c] for c in hq_bins}
        label_to_node["mq/lq"] = []
        for c in cluster_to_contig:
            if c not in hq_bins:
                label_to_node["mq/lq"] += list(cluster_to_contig[c])
    if centroids is not None:
        all_embs = tsne.fit_transform(np.concatenate((np.array(embs), np.array(centroids)), axis=0))
        centroids_2dim = all_embs[embs.shape[0] :]
        node_embeddings_2dim = all_embs[: embs.shape[0]]
    else:
        centroids_2dim = None
        node_embeddings_2dim = tsne.fit_transform(embs)
    return node_embeddings_2dim, centroids_2dim

def draw_nx_graph(adj, node_to_label, labels_to_node, basename, contig_sizes=None, node_titles=None, cluster_info={}):
    # draw graph with pybiz library, creates an HTML file with graph
    # del labels_to_node["NA"]
    from pyvis.network import Network

    labels_to_color = {l: colors[i % len(colors)] for i, l in enumerate(labels_to_node.keys())}
    labels_to_color["NA"] = "white"
    sorted(labels_to_node, key=lambda key: len(labels_to_node[key]), reverse=True)[: len(colors)]
    # node_labels to plot
    node_labels = {
        node: {
            "label": str(node) + ":" + str(node_to_label[node]),
            "color": labels_to_color[node_to_label[node]],
        }
        for node in node_to_label
    }
    if contig_sizes is not None:
        for n in node_labels:
            node_labels[n]["size"] = int(contig_sizes[n])

    if node_titles is not None:
        for n in node_labels:
            node_labels[n]["title"] = node_titles[n]
    #breakpoint()
    graph = nx.from_scipy_sparse_matrix(adj, parallel_edges=False, create_using=None, edge_attribute='weight')
    graph.remove_edges_from(nx.selfloop_edges(graph))
    nx.set_node_attributes(graph, node_labels)
    nodes_to_plot = [n for n in graph.nodes() if node_to_label[n] != "NA" and len(graph.edges(n)) > 0]

    net = Network(notebook=False, height="750px", width="100%")
    net.add_nodes(
        [int(n) for n in node_labels.keys() if n in nodes_to_plot],
        label=[node_labels[n]["label"] for n in node_labels  if n in nodes_to_plot],
        size=[node_labels[n].get("size", 100000)/100_000 for n in node_labels if n in nodes_to_plot],
        color=[node_labels[n]["color"] for n in node_labels if n in nodes_to_plot],
        title=[node_labels[n].get("title", f"{cluster_info.get(node_to_label[n])}") for n in node_labels if n in nodes_to_plot],
    )
    for u, v, a in graph.edges(data=True):
        if u != v:
            #if u not in net.get_nodes() or v not in net.get_nodes():
            #    breakpoint()
            weight = float(a["weight"].item())
            if weight != 1:
                net.add_edge(int(u), int(v), color="gray", title="reads weight: {}".format(weight))
            else:
                net.add_edge(int(u), int(v))

    #net.toggle_physics(False)
    net.show_buttons()
    print("saving graph to", basename)
    net.show("{}.html".format(basename))


def connected_components(graph, node_to_label, basename, min_elems=1):
    # explore connected components
    connected = [c for c in sorted(nx.connected_components(graph), key=len, reverse=True) if len(c) > min_elems]
    print("writing components to", basename + "_node_to_component.csv")
    write_components_file(connected, basename + "_node_to_component.csv")
    multiple_contigs = 0
    mixed_contigs = 0
    for group in connected:
        multiple_contigs += 1
        group_labels = [node_to_label[c] for c in group if "edge" in c]
        group_labels = set(group_labels)
        if len(group_labels) > 1:
            mixed_contigs += 1
            # print(group, group_labels)Process some integers.

    disconnected = [c for c in sorted(nx.connected_components(graph), key=len, reverse=True) if len(c) <= min_elems]
    for group in disconnected:
        for node in group:
            graph.remove_node(node)

    print("graph density:", nx.density(graph))
    print(">1", multiple_contigs)
    print("mixed groups", mixed_contigs)
    return connected, disconnected


def write_components_file(components, outpath, minsize=2):
    """Write file mapping each contig/node to a component ID (diff for each assembly)

    Args:
        components (list): List of connected components of a graph
        outpath (str): path to write file
        minsize: minimum number of elements of a component
    """
    with open(outpath, "w") as outfile:
        for ic, c in enumerate(components):
            if len(c) < minsize:
                continue
            for node in c:
                if "edge" in node:  # ignore read nodes
                    outfile.write(f"{node}\t{ic}\n")
